import collections
import dataclasses
import itertools
import math
import pprint
import re
from random import Random
from typing import Tuple, Iterator, NamedTuple, Set, AbstractSet, Union, Dict, \
    DefaultDict, Mapping, FrozenSet, Callable, List, TypeVar, Any, Optional

from randovania.game_description.assignment import PickupTarget
from randovania.game_description.game_description import calculate_interesting_resources, GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import Hint, HintType
from randovania.game_description.node import ResourceNode, Node
from randovania.game_description.requirements import RequirementList
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import ResourceInfo, CurrentResources
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world import World
from randovania.game_description.world_list import WorldList
from randovania.generator.filler.filler_library import UnableToGenerate, should_have_hint
from randovania.generator.generator_reach import GeneratorReach, collectable_resource_nodes, \
    advance_reach_with_possible_unsafe_resources, reach_with_all_safe_resources, \
    get_collectable_resource_nodes_of_reach, advance_to_with_reach_copy
from randovania.layout.available_locations import RandomizationMode
from randovania.resolver import debug
from randovania.resolver.random_lib import select_element_with_weight
from randovania.resolver.state import State

X = TypeVar("X")

_RESOURCES_WEIGHT_MULTIPLIER = 1
_INDICES_WEIGHT_MULTIPLIER = 1
_LOGBOOKS_WEIGHT_MULTIPLIER = 1
_VICTORY_WEIGHT = 1000
WeightedLocations = Dict[Tuple["PlayerState", PickupIndex], float]


@dataclasses.dataclass(frozen=True)
class FillerConfiguration:
    randomization_mode: RandomizationMode
    minimum_random_starting_items: int
    maximum_random_starting_items: int
    indices_to_exclude: FrozenSet[PickupIndex]


def _filter_not_in_dict(elements: Iterator[X],
                        dictionary: Dict[X, Any],
                        ) -> Set[X]:
    return set(elements) - set(dictionary.keys())


class UncollectedState(NamedTuple):
    indices: Set[PickupIndex]
    logbooks: Set[LogbookAsset]
    resources: Set[ResourceNode]

    @classmethod
    def from_reach(cls, reach: GeneratorReach) -> "UncollectedState":
        return UncollectedState(
            _filter_not_in_dict(reach.state.collected_pickup_indices, reach.state.patches.pickup_assignment),
            _filter_not_in_dict(reach.state.collected_scan_assets, reach.state.patches.hints),
            set(collectable_resource_nodes(reach.connected_nodes, reach))
        )

    def __sub__(self, other: "UncollectedState") -> "UncollectedState":
        return UncollectedState(
            self.indices - other.indices,
            self.logbooks - other.logbooks,
            self.resources - other.resources
        )


def find_node_with_resource(resource: ResourceInfo,
                            haystack: Iterator[Node],
                            ) -> ResourceNode:
    for node in haystack:
        if isinstance(node, ResourceNode) and node.resource() == resource:
            return node


def _calculate_reach_for_progression(reach: GeneratorReach,
                                     progression: PickupEntry,
                                     ) -> GeneratorReach:
    return advance_to_with_reach_copy(reach, reach.state.assign_pickup_resources(progression))


Action = Union[ResourceNode, PickupEntry]


def _resources_in_pickup(pickup: PickupEntry, current_resources: CurrentResources) -> FrozenSet[ResourceInfo]:
    resource_gain = pickup.resource_gain(current_resources)
    return frozenset(resource for resource, _ in resource_gain)


def _calculate_uncollected_index_weights(uncollected_indices: AbstractSet[PickupIndex],
                                         assigned_indices: AbstractSet[PickupIndex],
                                         seen_counts: Mapping[PickupIndex, int],
                                         indices_groups: List[Set[PickupIndex]],
                                         ) -> Dict[PickupIndex, float]:
    result = {}

    for indices in indices_groups:
        weight_from_collected_indices = math.sqrt(len(indices) / ((1 + len(assigned_indices & indices)) ** 2))

        for index in sorted(uncollected_indices & indices):
            weight_from_seen_count = min(10, seen_counts[index]) ** -2
            result[index] = weight_from_collected_indices * weight_from_seen_count
            # print(f"## {index} : {weight_from_collected_indices} ___ {weight_from_seen_count}")

    return result


def world_indices_for_mode(world: World, randomization_mode: RandomizationMode) -> Iterator[PickupIndex]:
    if randomization_mode is RandomizationMode.FULL:
        yield from world.pickup_indices
    elif randomization_mode is RandomizationMode.MAJOR_MINOR_SPLIT:
        yield from world.major_pickup_indices
    else:
        raise RuntimeError("Unknown randomization_mode: {}".format(randomization_mode))


def build_available_indices(world_list: WorldList, configuration: FillerConfiguration,
                            ) -> Tuple[List[Set[PickupIndex]], Set[PickupIndex]]:
    """
    Groups indices into separated groups, so each group can be weighted separately.
    """
    indices_groups = [
        set(world_indices_for_mode(world, configuration.randomization_mode)) - configuration.indices_to_exclude
        for world in world_list.worlds
    ]
    all_indices = set().union(*indices_groups)

    return indices_groups, all_indices


class PlayerState:
    index: int
    game: GameDescription
    pickups_left: List[PickupEntry]
    configuration: FillerConfiguration
    pickup_index_seen_count: DefaultDict[PickupIndex, int]
    scan_asset_seen_count: DefaultDict[LogbookAsset, int]
    scan_asset_initial_pickups: Dict[LogbookAsset, FrozenSet[PickupIndex]]
    _unfiltered_potential_actions: List[Action]
    num_random_starting_items_placed: int
    num_assigned_pickups: int

    def __init__(self,
                 index: int,
                 game: GameDescription,
                 initial_state: State,
                 pickups_left: List[PickupEntry],
                 configuration: FillerConfiguration,
                 ):
        self.index = index
        self.game = game
        self.reach = advance_reach_with_possible_unsafe_resources(reach_with_all_safe_resources(game, initial_state))
        self.pickups_left = pickups_left
        self.configuration = configuration

        self.pickup_index_seen_count = collections.defaultdict(int)
        self.scan_asset_seen_count = collections.defaultdict(int)
        self.scan_asset_initial_pickups = {}
        self.num_random_starting_items_placed = 0
        self.num_assigned_pickups = 0
        self.num_actions = 0
        self.indices_groups, self.all_indices = build_available_indices(game.world_list, configuration)

    def __repr__(self):
        return f"Player {self.index}"

    def update_for_new_state(self):
        self._advance_pickup_index_seen_count()
        self._advance_scan_asset_seen_count()
        self._calculate_potential_actions()

    def _advance_pickup_index_seen_count(self):
        for pickup_index in self.reach.state.collected_pickup_indices:
            self.pickup_index_seen_count[pickup_index] += 1

        print_new_resources(self.game, self.reach, self.pickup_index_seen_count, "Pickup Index")

    def _advance_scan_asset_seen_count(self):
        for scan_asset in self.reach.state.collected_scan_assets:
            self.scan_asset_seen_count[scan_asset] += 1
            if self.scan_asset_seen_count[scan_asset] == 1:
                self.scan_asset_initial_pickups[scan_asset] = frozenset(self.reach.state.collected_pickup_indices)

        print_new_resources(self.game, self.reach, self.scan_asset_seen_count, "Scan Asset")

    def _calculate_potential_actions(self):
        progression_pickups = _calculate_progression_pickups(self.pickups_left, self.reach)
        print_retcon_loop_start(self.game, self.pickups_left, self.reach, self.index)

        uncollected_resource_nodes = get_collectable_resource_nodes_of_reach(self.reach)
        self._unfiltered_potential_actions = list(progression_pickups)
        self._unfiltered_potential_actions.extend(uncollected_resource_nodes)

    def potential_actions(self, num_available_indices: int) -> List[Action]:
        num_available_indices += (self.configuration.maximum_random_starting_items
                                  - self.num_random_starting_items_placed)
        return [
            action
            for action in self._unfiltered_potential_actions
            if not isinstance(action, PickupEntry) or _items_for_pickup(action) <= num_available_indices
        ]

    def weighted_potential_actions(self, status_update: Callable[[str], None], num_available_indices: int,
                                   ) -> Dict[Action, float]:
        """
        Weights all potential actions based on current criteria.
        :param status_update:
        :param num_available_indices: The number of indices available for placement.
        :return:
        """
        actions_weights: Dict[Action, float] = {}
        current_uncollected = UncollectedState.from_reach(self.reach)

        actions = self.potential_actions(num_available_indices)
        options_considered = 0

        def update_for_option():
            nonlocal options_considered
            options_considered += 1
            status_update("Checked {} of {} options.".format(options_considered, len(actions)))

        for action in actions:
            if isinstance(action, PickupEntry):
                base_weight = _calculate_weights_for(_calculate_reach_for_progression(self.reach, action),
                                                     current_uncollected,
                                                     action.name)
                weight = base_weight * action.probability_multiplier + action.probability_offset

            else:
                weight = _calculate_weights_for(
                    advance_to_with_reach_copy(self.reach, self.reach.state.act_on_node(action)),
                    current_uncollected,
                    action.name)

            actions_weights[action] = weight
            update_for_option()

        if debug.debug_level() > 1:
            for action, weight in actions_weights.items():
                print("({}) {} - {}".format(type(action).__name__, action.name, weight))

        return actions_weights

    def victory_condition_satisfied(self):
        return self.game.victory_condition.satisfied(self.reach.state.resources, self.reach.state.energy)

    def assign_pickup(self, pickup_index: PickupIndex, target: PickupTarget):
        self.num_assigned_pickups += 1
        self.reach.state.patches = self.reach.state.patches.assign_new_pickups([
            (pickup_index, target),
        ])

    def current_state_report(self) -> str:
        state = UncollectedState.from_reach(self.reach)
        pickups_by_name_and_quantity = collections.defaultdict(int)

        _KEY_MATCH = re.compile(r"Key (\d+)")
        for pickup in self.pickups_left:
            pickups_by_name_and_quantity[_KEY_MATCH.sub("Key", pickup.name)] += 1

        to_progress = {_KEY_MATCH.sub("Key", resource.long_name)
                       for resource in interesting_resources_for_reach(self.reach)
                       if resource.resource_type == ResourceType.ITEM}

        return ("At {0} after {1} actions and {2} pickups, with {3} collected locations.\n\n"
                "Pickups still available: {4}\n\nResources to progress: {5}").format(
            self.game.world_list.node_name(self.reach.state.node, with_world=True, distinguish_dark_aether=True),
            self.num_actions,
            self.num_assigned_pickups,
            len(state.indices),
            ", ".join(name if quantity == 1 else f"{name} x{quantity}"
                      for name, quantity in sorted(pickups_by_name_and_quantity.items())),
            ", ".join(sorted(to_progress)),
        )


def _get_next_player(rng: Random, player_states: List[PlayerState], num_indices: int) -> Optional[PlayerState]:
    """
    Gets the next player a pickup should be placed for.
    :param rng:
    :param player_states:
    :param num_indices: The number of indices av
    :return:
    """
    all_uncollected: Dict[PlayerState, UncollectedState] = {
        player_state: UncollectedState.from_reach(player_state.reach)
        for player_state in player_states
    }

    max_actions = max(player_state.num_actions for player_state in player_states)
    max_uncollected = max(len(uncollected.indices) for uncollected in all_uncollected.values())

    def _calculate_weight(player: PlayerState) -> float:
        return 1 + (max_actions - player.num_actions) * (max_uncollected - len(all_uncollected[player].indices))

    weighted_players = {
        player_state: _calculate_weight(player_state)
        for player_state in player_states
        if not player_state.victory_condition_satisfied() and player_state.potential_actions(num_indices)
    }
    if weighted_players:
        if debug.debug_level() > 1:
            print(f">>>>> Player Weights: {weighted_players}")

        return select_element_with_weight(weighted_players, rng)
    else:
        if all(player_state.victory_condition_satisfied() for player_state in player_states):
            debug.debug_print("Finished because we can win")
            return None
        else:
            total_actions = sum(player_state.num_actions for player_state in player_states)
            raise UnableToGenerate(f"No players with possible actions after {total_actions} total actions.")


def retcon_playthrough_filler(rng: Random,
                              player_states: List[PlayerState],
                              status_update: Callable[[str], None],
                              ) -> Tuple[Dict[PlayerState, GamePatches], Tuple[str, ...]]:
    """
    Runs the retcon logic.
    :param rng:
    :param player_states:
    :param status_update:
    :return: A GamePatches for each player and a sequence of placed items.
    """
    debug.debug_print("{}\nRetcon filler started with major items:\n{}".format(
        "*" * 100,
        "\n".join(
            "Player {}: {}".format(
                player_state.index,
                pprint.pformat({
                    item.name: player_state.pickups_left.count(item)
                    for item in sorted(set(player_state.pickups_left), key=lambda item: item.name)
                })
            )
            for player_state in player_states
        )
    ))
    last_message = "Starting."

    def action_report(message: str):
        status_update("{} {}".format(last_message, message))

    for player_state in player_states:
        player_state.update_for_new_state()

    actions_log = []

    while True:
        all_locations_weighted = _calculate_all_pickup_indices_weight(player_states)
        current_player = _get_next_player(rng, player_states, len(all_locations_weighted))
        if current_player is None:
            break

        weighted_actions = current_player.weighted_potential_actions(action_report, len(all_locations_weighted))
        try:
            action = select_element_with_weight(weighted_actions, rng=rng)
        except StopIteration:
            # All actions had weight 0. Select one randomly instead.
            # No need to check if potential_actions is empty, _get_next_player only return players with actions
            action = rng.choice(list(weighted_actions.keys()))

        if isinstance(action, PickupEntry):
            log_entry = _assign_pickup_somewhere(action, current_player, player_states, rng, all_locations_weighted)
            actions_log.append(log_entry)
            debug.debug_print(f"\n>>>> {log_entry}")

            # TODO: this item is potentially dangerous and we should remove the invalidated paths
            current_player.pickups_left.remove(action)
            current_player.num_actions += 1

            count_pickups_left = sum(len(player_state.pickups_left) for player_state in player_states)
            last_message = "{} items left.".format(count_pickups_left)
            status_update(last_message)

        else:
            last_message = "Triggered an event out of {} options.".format(len(weighted_actions))
            status_update(last_message)
            debug_print_collect_event(action, current_player.game)

            # This action is potentially dangerous. Use `act_on` to remove invalid paths
            current_player.reach.act_on(action)

        current_player.reach = advance_reach_with_possible_unsafe_resources(current_player.reach)
        current_player.update_for_new_state()

    all_patches = {player_state: player_state.reach.state.patches for player_state in player_states}
    return all_patches, tuple(actions_log)


def _assign_pickup_somewhere(action: PickupEntry,
                             current_player: PlayerState,
                             player_states: List[PlayerState],
                             rng: Random,
                             all_locations_weighted: WeightedLocations,
                             ) -> str:
    """
    Assigns a PickupEntry to a free, collected PickupIndex or as a starting item.
    :param action:
    :param current_player:
    :param player_states:
    :param rng:
    :return:
    """
    assert action in current_player.pickups_left

    if all_locations_weighted and (current_player.num_random_starting_items_placed
                                   >= current_player.configuration.minimum_random_starting_items):

        index_owner_state, pickup_index = select_element_with_weight(all_locations_weighted, rng)
        index_owner_state.assign_pickup(pickup_index, PickupTarget(action, current_player.index))

        # Place a hint for the new item
        hint_location = _calculate_hint_location_for_action(
            action,
            UncollectedState.from_reach(index_owner_state.reach),
            pickup_index,
            rng,
            index_owner_state.scan_asset_initial_pickups,
        )
        if hint_location is not None:
            index_owner_state.reach.state.patches = index_owner_state.reach.state.patches.assign_hint(
                hint_location, Hint(HintType.LOCATION, None, pickup_index))

        if pickup_index in index_owner_state.reach.state.collected_pickup_indices:
            current_player.reach.advance_to(current_player.reach.state.assign_pickup_resources(action))
        else:
            # FIXME: isn't that condition always true?
            pass

        spoiler_entry = pickup_placement_spoiler_entry(current_player.index, action, index_owner_state.game,
                                                       pickup_index, hint_location, index_owner_state.index,
                                                       len(player_states) > 1)

    else:
        current_player.num_random_starting_items_placed += 1
        if (current_player.num_random_starting_items_placed
                > current_player.configuration.maximum_random_starting_items):
            raise UnableToGenerate("Attempting to place more extra starting items than the number allowed.")

        spoiler_entry = f"{action.name} as starting item"
        if len(player_states) > 1:
            spoiler_entry += f" for Player {current_player.index + 1}"
        current_player.reach.advance_to(current_player.reach.state.assign_pickup_to_starting_items(action))

    return spoiler_entry


def _calculate_all_pickup_indices_weight(player_states: List[PlayerState]) -> WeightedLocations:
    all_weights = {}

    total_assigned_pickups = sum(player_state.num_assigned_pickups for player_state in player_states)

    # print("================ WEIGHTS! ==================")
    for player_state in player_states:
        delta = (total_assigned_pickups - player_state.num_assigned_pickups)
        player_weight = 1 + delta

        # print(f"** Player {player_state.index} -- {player_weight}")

        pickup_index_weights = _calculate_uncollected_index_weights(
            player_state.all_indices & UncollectedState.from_reach(player_state.reach).indices,
            set(player_state.reach.state.patches.pickup_assignment),
            player_state.pickup_index_seen_count,
            player_state.indices_groups,
        )
        for pickup_index, weight in pickup_index_weights.items():
            all_weights[(player_state, pickup_index)] = weight * player_weight

    # for (player_state, pickup_index), weight in all_weights.items():
    #     print(f"> {player_state.index} - {pickup_index}: {weight}")
    # print("============================================")

    return all_weights


def _calculate_hint_location_for_action(action: PickupEntry,
                                        current_uncollected: UncollectedState,
                                        pickup_index: PickupIndex,
                                        rng: Random,
                                        scan_asset_initial_pickups: Dict[LogbookAsset, FrozenSet[PickupIndex]],
                                        ) -> Optional[LogbookAsset]:
    """
    Calculates where a hint for the given action should be placed.
    :return: A LogbookAsset to use, or None if no hint should be placed.
    """
    if should_have_hint(action.item_category):
        potential_hint_locations = [
            logbook_asset
            for logbook_asset in current_uncollected.logbooks
            if pickup_index not in scan_asset_initial_pickups[logbook_asset]
        ]
        if potential_hint_locations:
            return rng.choice(sorted(potential_hint_locations))
    return None


def interesting_resources_for_reach(reach: GeneratorReach) -> FrozenSet[ResourceInfo]:
    satisfiable_requirements: FrozenSet[RequirementList] = frozenset(itertools.chain.from_iterable(
        requirements.alternatives
        for requirements in reach.unreachable_nodes_with_requirements().values()
    ))
    return calculate_interesting_resources(
        satisfiable_requirements,
        reach.state.resources,
        reach.state.energy,
        reach.state.resource_database
    )


def _calculate_progression_pickups(pickups_left: Iterator[PickupEntry],
                                   reach: GeneratorReach,
                                   ) -> Tuple[PickupEntry, ...]:
    interesting_resources = interesting_resources_for_reach(reach)
    progression_pickups = []

    for pickup in pickups_left:
        if pickup in progression_pickups:
            continue
        if _resources_in_pickup(pickup, reach.state.resources).intersection(interesting_resources):
            progression_pickups.append(pickup)

    return tuple(progression_pickups)


def _calculate_weights_for(potential_reach: GeneratorReach,
                           current_uncollected: UncollectedState,
                           name: str
                           ) -> float:
    if potential_reach.game.victory_condition.satisfied(potential_reach.state.resources,
                                                        potential_reach.state.energy):
        return _VICTORY_WEIGHT

    potential_uncollected = UncollectedState.from_reach(potential_reach) - current_uncollected
    return sum((
        _RESOURCES_WEIGHT_MULTIPLIER * int(bool(potential_uncollected.resources)),
        _INDICES_WEIGHT_MULTIPLIER * int(bool(potential_uncollected.indices)),
        _LOGBOOKS_WEIGHT_MULTIPLIER * int(bool(potential_uncollected.logbooks)),
    ))


def _items_for_pickup(pickup: PickupEntry) -> int:
    return 1


def debug_print_collect_event(action, game):
    if debug.debug_level() > 0:
        print("\n--> Collecting {}".format(game.world_list.node_name(action, with_world=True)))


def print_retcon_loop_start(game: GameDescription,
                            pickups_left: Iterator[PickupEntry],
                            reach: GeneratorReach,
                            player_index: int,
                            ):
    if debug.debug_level() > 0:
        current_uncollected = UncollectedState.from_reach(reach)
        if debug.debug_level() > 1:
            extra = ", pickups_left: {}".format(sorted(set(pickup.name for pickup in pickups_left)))
        else:
            extra = ""

        print("\n\n===============================")
        print("\n>>> Player {}: From {}, {} open pickup indices, {} open resources{}".format(
            player_index,
            game.world_list.node_name(reach.state.node, with_world=True),
            len(current_uncollected.indices),
            len(current_uncollected.resources),
            extra
        ))

        if debug.debug_level() > 2:
            print("\nCurrent reach:")
            for node in reach.nodes:
                print("[{!s:>5}, {!s:>5}] {}".format(reach.is_reachable_node(node), reach.is_safe_node(node),
                                                     game.world_list.node_name(node)))


def pickup_placement_spoiler_entry(owner_index: int, action: PickupEntry, game: GameDescription,
                                   pickup_index: PickupIndex, hint: Optional[LogbookAsset],
                                   player_index: int, add_indices: bool) -> str:
    world_list = game.world_list
    if hint is not None:
        hint_string = " with hint at {}".format(
            world_list.node_name(find_node_with_resource(hint, world_list.all_nodes),
                                 with_world=True, distinguish_dark_aether=True))
    else:
        hint_string = ""

    pickup_node = find_node_with_resource(pickup_index, world_list.all_nodes)
    return "{4}{0} at {3}{1}{2}".format(
        action.name,
        world_list.node_name(pickup_node, with_world=True, distinguish_dark_aether=True),
        hint_string,
        f"player {player_index + 1}'s " if add_indices else "",
        f"Player {owner_index + 1}'s " if add_indices else "",
    )


def print_new_resources(game: GameDescription,
                        reach: GeneratorReach,
                        seen_count: Dict[ResourceInfo, int],
                        label: str,
                        ):
    world_list = game.world_list
    if debug.debug_level() > 1:
        for index, count in seen_count.items():
            if count == 1:
                node = find_node_with_resource(index, world_list.all_nodes)
                print("-> New {}: {}".format(label, world_list.node_name(node, with_world=True)))
        print("")



from typing import Set, Callable, TypeVar

from randovania.game_description.game_description import GameDescription
from randovania.game_description.requirements import Requirement
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.layout.trick_level import LayoutTrickLevel

T = TypeVar("T")


def _process_connections(game: GameDescription, process: Callable[[Requirement, Set[T]], None]) -> Set[T]:
    result = set()

    for dock_weaknesses in game.dock_weakness_database:
        for dock_weakness in dock_weaknesses:
            process(dock_weakness.requirement, result)

    for area in game.world_list.all_areas:
        for _, _, requirement in area.all_connections:
            process(requirement, result)

    return result


def difficulties_for_trick(game: GameDescription, trick: TrickResourceInfo) -> Set[LayoutTrickLevel]:
    def process(req: Requirement, result: Set[LayoutTrickLevel]):
        for resource_requirement in req.iterate_resource_requirements():
            if resource_requirement.resource == trick:
                result.add(LayoutTrickLevel.from_number(resource_requirement.amount))

    return _process_connections(game, process)


def used_tricks(game: GameDescription) -> Set[TrickResourceInfo]:
    def process(req: Requirement, result: Set[TrickResourceInfo]):
        for resource_requirement in req.iterate_resource_requirements():
            if resource_requirement.resource.resource_type == ResourceType.TRICK:
                result.add(resource_requirement.resource)

    return _process_connections(game, process)



import io
import json
from pathlib import Path

import pytest

from randovania import get_data_path
from randovania.games.game import RandovaniaGame
from randovania.games.prime import binary_data


def test_simple_round_trip():
    sample_data = {
        "game": "prime2",
        "resource_database": {
            "items": [],
            "energy_tank_item_index": 0,
            "item_percentage_index": 0,
            "multiworld_magic_item_index": 0,
            "events": [],
            "tricks": [],
            "damage": [],
            "versions": [],
            "misc": [],
            "requirement_template": {},
        },
        "game_specific": {
            "energy_per_tank": 100.0,
            "safe_zone_heal_per_second": 1.0,
            "beam_configurations": [],
            "dangerous_energy_tank": True,
        },
        "starting_location": {
            "world_asset_id": 1006255871,
            "area_asset_id": 1655756413
        },
        "initial_states": {
            "Default": [
            ]
        },
        "victory_condition": {
            "type": "and",
            "data": []
        },
        "dock_weakness_database": {
            "door": [],
            "portal": [],
            "morph_ball": [],
        },
        "worlds": [],
    }

    b = io.BytesIO()
    binary_data.encode(sample_data, b)

    b.seek(0)
    decoded = binary_data.decode(b)

    assert decoded == sample_data


def test_complex_encode(test_files_dir):
    with test_files_dir.joinpath("prime_data_as_json.json").open("r") as data_file:
        data = json.load(data_file)

    b = io.BytesIO()

    # Run
    binary_data.encode(data, b)
    # Whenever the file format changes, we can use the following line to force update our test file
    # test_files_dir.joinpath("prime_data_as_binary.bin").write_bytes(b.getvalue()); assert False

    # Assert
    assert test_files_dir.joinpath("prime_data_as_binary.bin").read_bytes() == b.getvalue()


def test_complex_decode(test_files_dir):
    # Run
    decoded_data = binary_data.decode_file_path(Path(test_files_dir.joinpath("prime_data_as_binary.bin")))

    # Assert
    with test_files_dir.joinpath("prime_data_as_json.json").open("r") as data_file:
        saved_data = json.load(data_file)

    assert decoded_data == saved_data


@pytest.mark.parametrize("game", RandovaniaGame)
def test_full_file_round_trip(game):
    # Setup
    json_database_path = get_data_path().joinpath("json_data", f"{game.value}.json")
    if not json_database_path.exists():
        pytest.skip("Missing database")

    with json_database_path.open("r") as json_file:
        original_data = json.load(json_file)

    # Run 1
    output_io = io.BytesIO()
    binary_data.encode(original_data, output_io)

    # Run 2
    output_io.seek(0)
    final_data = binary_data.decode(output_io)

    # Assert
    assert final_data == original_data


def _comparable_dict(value):
    if isinstance(value, dict):
        return [
            (key, _comparable_dict(item))
            for key, item in value.items()
        ]

    if isinstance(value, list):
        return [_comparable_dict(item) for item in value]

    return value


@pytest.mark.skipif(
    not get_data_path().joinpath("json_data", "prime2.json").is_file(),
    reason="Missing prime2.json"
)
def test_full_data_encode_is_equal():
    # The prime2.json may be missing if we're running using a Pyinstaller binary
    # Setup
    json_database_file = get_data_path().joinpath("json_data", "prime2.json")

    with json_database_file.open("r") as open_file:
        json_database = json.load(open_file)

    b = io.BytesIO()
    binary_data.encode(json_database, b)

    b.seek(0)
    decoded_database = binary_data.decode(b)

    # Run
    assert json_database == decoded_database

    comparable_json = _comparable_dict(json_database)
    comparable_binary = _comparable_dict(decoded_database)
    assert comparable_json == comparable_binary


@pytest.mark.parametrize("req", [
    {"type": "or", "data": []},
    {"type": "and", "data": []},
    {"type": "resource", "data": {"type": 2, "index": 5, "amount": 7, "negate": True}},
    {"type": "template", "data": "Example Template"},
])
def test_encode_requirement_simple(req):
    # Run
    encoded = binary_data.ConstructRequirement.build(req)
    decoded = binary_data._convert_to_raw_python(binary_data.ConstructRequirement.parse(encoded))

    # Assert
    assert req == decoded


def test_encode_requirement_complex():
    # Setup
    req = {
        "type": "and",
        "data": [
            {"type": "or", "data": []},
            {"type": "and", "data": []},
            {"type": "or", "data": []},
            {"type": "resource", "data": {"type": 2, "index": 5, "amount": 7, "negate": True}},
            {"type": "or", "data": []},
            {"type": "template", "data": "Example Template"},
        ]
    }

    # Run
    encoded = binary_data.ConstructRequirement.build(req)
    decoded = binary_data._convert_to_raw_python(binary_data.ConstructRequirement.parse(encoded))

    # Assert
    assert req == decoded


def test_encode_resource_database():
    # Setup
    resource_database = {
        "items": [],
        "energy_tank_item_index": 0,
        "item_percentage_index": 0,
        "multiworld_magic_item_index": 0,
        "events": [],
        "tricks": [],
        "damage": [],
        "versions": [],
        "misc": [],
        "requirement_template": {
            "Foo": {
                "type": "or",
                "data": []
            }
        },
    }
    resource_database["requirement_template"] = list(resource_database["requirement_template"].items())

    # Run
    encoded = binary_data.ConstructResourceDatabase.build(resource_database)

    # Assert
    assert encoded == b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01Foo\x00\x02\x00'



import copy
from pathlib import Path
from typing import TypeVar, BinaryIO, Dict, Any

import construct
from construct import Struct, Int32ub, Const, CString, Byte, Rebuild, Float32b, Flag, \
    Short, PrefixedArray, Array, Switch, If, VarInt, Sequence, Float64b

from randovania.game_description.node import LoreType
from randovania.games.game import RandovaniaGame

X = TypeVar('X')
current_format_version = 8

_IMPOSSIBLE_SET = {"type": "or", "data": []}


def _convert_to_raw_python(value) -> Any:
    if isinstance(value, construct.ListContainer):
        return [
            _convert_to_raw_python(item)
            for item in value
        ]

    if isinstance(value, construct.Container):
        return {
            key: _convert_to_raw_python(item)
            for key, item in value.items()
            if not key.startswith("_")
        }

    if isinstance(value, construct.EnumIntegerString):
        return str(value)

    return value


def decode(binary_io: BinaryIO) -> Dict:
    decoded = _convert_to_raw_python(ConstructGame.parse_stream(binary_io))

    decoded.pop("format_version")
    decoded.pop("magic_number")
    decoded["initial_states"] = dict(decoded["initial_states"])
    decoded["resource_database"]["requirement_template"] = dict(decoded["resource_database"]["requirement_template"])

    for world in decoded["worlds"]:
        for area in world["areas"]:

            for node in area["nodes"]:
                data = node.pop("data")
                if data is not None:
                    for key, value in data.items():
                        node[key] = value
                node["connections"] = {}

            for i, connections in enumerate(area.pop("connections")):
                node = area["nodes"][i]

                j = 0
                for connection in connections:
                    if j == i:
                        j += 1

                    if connection != _IMPOSSIBLE_SET:
                        node["connections"][area["nodes"][j]["name"]] = connection
                    j += 1

    fields = [
        "game",
        "resource_database",
        "game_specific",
        "starting_location",
        "initial_states",
        "victory_condition",
        "dock_weakness_database",
        "worlds",
    ]
    result = {
        field_name: decoded.pop(field_name)
        for field_name in fields
    }
    if decoded:
        raise ValueError(f"Unexpected fields remaining in data: {list(decoded.keys())}")

    return result


def decode_file_path(binary_file_path: Path) -> Dict:
    with binary_file_path.open("rb") as binary_io:
        return decode(binary_io)


def encode(original_data: Dict, x: BinaryIO) -> None:
    data = copy.deepcopy(original_data)

    for world in data["worlds"]:
        for area in world["areas"]:
            area["connections"] = [
                [
                    origin["connections"].get(target["name"], _IMPOSSIBLE_SET)
                    for target in area["nodes"]
                    if origin != target
                ]
                for origin in area["nodes"]
            ]

            for i, node in enumerate(area["nodes"]):
                node.pop("connections")
                area["nodes"][i] = {
                    "name": node.pop("name"),
                    "heal": node.pop("heal"),
                    "coordinates": node.pop("coordinates"),
                    "node_type": node.pop("node_type"),
                    "data": node,
                }

    data["resource_database"]["requirement_template"] = list(data["resource_database"]["requirement_template"].items())
    data["initial_states"] = list(data["initial_states"].items())

    ConstructGame.build_stream(data, x)

    # Resource Info database
    data.pop("game")
    data.pop("resource_database")
    data.pop("game_specific")
    data.pop("dock_weakness_database")
    data.pop("worlds")
    data.pop("victory_condition")
    data.pop("starting_location")
    data.pop("initial_states")

    if data:
        raise ValueError(f"Unexpected fields remaining in data: {list(data.keys())}")


def OptionalValue(subcon):
    return construct.FocusedSeq(
        "value",
        present=Rebuild(Flag, construct.this.value != None),
        value=If(construct.this.present, subcon),
    )


def _build_resource_info(**kwargs):
    return Struct(
        index=VarInt,
        long_name=CString("utf8"),
        short_name=CString("utf8"),
        **kwargs,
    )


ConstructResourceInfo = _build_resource_info()

ConstructItemResourceInfo = _build_resource_info(
    max_capacity=Int32ub,
    extra=OptionalValue(Int32ub),
)

ConstructTrickResourceInfo = _build_resource_info(
    description=CString("utf8"),
)

ConstructDamageResourceInfo = _build_resource_info(
    reductions=PrefixedArray(VarInt, Struct(
        index=VarInt,
        multiplier=Float32b,
    ))
)


ConstructResourceRequirement = Struct(
    type=Byte,
    index=VarInt,
    amount=Short,
    negate=Flag,
)

requirement_type_map = {
    "resource": ConstructResourceRequirement,
    "template": CString("utf8"),
}

ConstructRequirement = Struct(
    type=construct.Enum(Byte, resource=0, **{"and": 1, "or": 2}, template=3),
    data=Switch(lambda this: this.type, requirement_type_map)
)
requirement_type_map["and"] = PrefixedArray(VarInt, ConstructRequirement)
requirement_type_map["or"] = PrefixedArray(VarInt, ConstructRequirement)

ConstructDockWeakness = Struct(
    index=VarInt,
    name=CString("utf8"),
    is_blast_door=Flag,
    requirement=ConstructRequirement,
)

ConstructResourceDatabase = Struct(
    items=PrefixedArray(VarInt, ConstructItemResourceInfo),
    energy_tank_item_index=VarInt,
    item_percentage_index=VarInt,
    multiworld_magic_item_index=VarInt,
    events=PrefixedArray(VarInt, ConstructResourceInfo),
    tricks=PrefixedArray(VarInt, ConstructTrickResourceInfo),
    damage=PrefixedArray(VarInt, ConstructDamageResourceInfo),
    versions=PrefixedArray(VarInt, ConstructResourceInfo),
    misc=PrefixedArray(VarInt, ConstructResourceInfo),
    requirement_template=PrefixedArray(VarInt, Sequence(CString("utf8"), ConstructRequirement))
)

ConstructEchoesBeamConfiguration = Struct(
    item_index=VarInt,
    ammo_a=OptionalValue(Byte),
    ammo_b=OptionalValue(Byte),
    uncharged_cost=Byte,
    charged_cost=Byte,
    combo_missile_cost=Byte,
    combo_ammo_cost=Byte,
)

ConstructEchoesGameSpecific = Struct(
    energy_per_tank=Float32b,
    safe_zone_heal_per_second=Float32b,
    beam_configurations=PrefixedArray(VarInt, ConstructEchoesBeamConfiguration),
    dangerous_energy_tank=Flag,
)

ConstructResourceGain = Struct(
    resource_type=Byte,
    resource_index=VarInt,
    amount=VarInt,
)

ConstructLoreType = construct.Enum(Byte, **{lore_type.value: i for i, lore_type in enumerate(LoreType)})

ConstructNodeCoordinates = Struct(
    x=Float64b,
    y=Float64b,
    z=Float64b,
)

ConstructNode = Struct(
    name=CString("utf8"),
    heal=Flag,
    coordinates=OptionalValue(ConstructNodeCoordinates),
    node_type=construct.Enum(Byte, generic=0, dock=1, pickup=2, teleporter=3, event=4, translator_gate=5,
                             logbook=6, player_ship=7),
    data=Switch(
        lambda this: this.node_type,
        {
            "dock": Struct(
                dock_index=Byte,
                connected_area_asset_id=VarInt,
                connected_dock_index=Byte,
                dock_type=Byte,
                dock_weakness_index=VarInt,
                _=Const(b"\x00\x00\x00"),
            ),
            "pickup": Struct(
                pickup_index=VarInt,
                major_location=Flag,
            ),
            "teleporter": Struct(
                destination_world_asset_id=VarInt,
                destination_area_asset_id=VarInt,
                teleporter_instance_id=OptionalValue(VarInt),
                scan_asset_id=OptionalValue(VarInt),
                keep_name_when_vanilla=Flag,
                editable=Flag,
            ),
            "event": Struct(
                event_index=VarInt,
            ),
            "translator_gate": Struct(
                gate_index=VarInt,
            ),
            "logbook": Struct(
                string_asset_id=VarInt,
                lore_type=ConstructLoreType,
                extra=VarInt,
            ),
            "player_ship": Struct(
                is_unlocked=ConstructRequirement,
            )
        }
    )
)

ConstructArea = Struct(
    name=CString("utf8"),
    in_dark_aether=Flag,
    asset_id=VarInt,
    _node_count=Rebuild(VarInt, construct.len_(construct.this.nodes)),
    default_node_index=OptionalValue(VarInt),
    valid_starting_location=Flag,
    nodes=Array(lambda this: this._node_count, ConstructNode),
    connections=Array(
        lambda this: this._node_count,
        Array(lambda this: this._node_count - 1, ConstructRequirement)
    )
)

ConstructWorld = Struct(
    name=CString("utf8"),
    dark_name=OptionalValue(CString("utf8")),
    asset_id=VarInt,
    areas=PrefixedArray(VarInt, ConstructArea),
)

ConstructGameEnum = construct.Enum(Byte, **{enum_item.value: i for i, enum_item in enumerate(RandovaniaGame)})

game_specific_map = {
    RandovaniaGame.PRIME1.value: Struct(),
    RandovaniaGame.PRIME2.value: ConstructEchoesGameSpecific,
    RandovaniaGame.PRIME3.value: Struct(),
}

ConstructGame = Struct(
    magic_number=Const(b"Req."),
    format_version=Const(current_format_version, Int32ub),
    game=ConstructGameEnum,
    resource_database=ConstructResourceDatabase,
    game_specific=Switch(lambda this: this.game, game_specific_map),
    dock_weakness_database=Struct(
        door=PrefixedArray(VarInt, ConstructDockWeakness),
        portal=PrefixedArray(VarInt, ConstructDockWeakness),
        morph_ball=PrefixedArray(VarInt, ConstructDockWeakness),
    ),
    victory_condition=ConstructRequirement,
    starting_location=Struct(
        world_asset_id=VarInt,
        area_asset_id=VarInt,
    ),
    initial_states=PrefixedArray(VarInt, construct.Sequence(CString("utf8"),
                                                            PrefixedArray(VarInt, ConstructResourceGain))),
    worlds=PrefixedArray(VarInt, ConstructWorld),
)



from typing import List, Optional, Dict

import dolphin_memory_engine

from randovania.game_connection.backend_choice import GameBackendChoice
from randovania.game_connection.connection_backend import ConnectionBackend, MemoryOperation, MemoryOperationException
from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_description.world import World

MEM1_START = 0x80000000
MEM1_END = 0x81800000


def _validate_range(address: int, size: int):
    if address < MEM1_START or address + size > MEM1_END:
        raise MemoryOperationException(
            f"Range {address:x} -> {address + size:x} is outside of the GameCube memory range.")


class DolphinBackend(ConnectionBackend):
    _world: Optional[World] = None

    def __init__(self):
        super().__init__()
        self.dolphin = dolphin_memory_engine

    @property
    def lock_identifier(self) -> Optional[str]:
        return "randovania-dolphin-backend"

    @property
    def backend_choice(self) -> GameBackendChoice:
        return GameBackendChoice.DOLPHIN

    # Game Backend Stuff
    def _memory_operation(self, op: MemoryOperation, pointers: Dict[int, Optional[int]]) -> Optional[bytes]:
        op.validate_byte_sizes()

        address = op.address
        if op.offset is not None:
            if address not in pointers:
                raise MemoryOperationException(f"Invalid op: {address} is not in pointers")
            address = pointers[address] + op.offset

        _validate_range(address, op.byte_count)

        if not self.dolphin.is_hooked():
            raise MemoryOperationException("Lost connection do Dolphin")

        result = None
        if op.read_byte_count is not None:
            result = self.dolphin.read_bytes(address, op.read_byte_count)

        if op.write_bytes is not None:
            self.dolphin.write_bytes(address, op.write_bytes)
            self.logger.debug(f"Wrote {op.write_bytes.hex()} to {address:x}")

        return result

    async def _perform_memory_operations(self, ops: List[MemoryOperation]) -> Dict[MemoryOperation, bytes]:
        pointers_to_read = set()
        for op in ops:
            if op.offset is not None:
                pointers_to_read.add(op.address)

        pointers = {}
        for pointer in pointers_to_read:
            if not self.dolphin.is_hooked():
                raise MemoryOperationException("Lost connection do Dolphin")

            try:
                pointers[pointer] = self.dolphin.follow_pointers(pointer, [0])
            except RuntimeError:
                self.logger.debug(f"Failed to read a valid pointer from {pointer:x}")
                self._test_still_hooked()

            if not self.dolphin.is_hooked():
                raise MemoryOperationException("Lost connection do Dolphin")

        result = {}
        for op in ops:
            op_result = self._memory_operation(op, pointers)
            if op_result is not None:
                result[op] = op_result
        return result

    def _ensure_hooked(self) -> bool:
        if not self.dolphin.is_hooked():
            self.patches = None
            self.dolphin.hook()

        return not self.dolphin.is_hooked()

    def _test_still_hooked(self):
        try:
            if len(self.dolphin.read_bytes(0x0, 4)) != 4:
                raise RuntimeError("Dolphin hook didn't read the correct byte count")
        except RuntimeError as e:
            self.logger.warning(f"Test read for Dolphin hook didn't work: {e}")
            self.dolphin.un_hook()

    async def update(self, dt: float):
        if not self._enabled:
            return

        if self._ensure_hooked():
            return

        if not await self._identify_game():
            self._test_still_hooked()
            return

        await self._interact_with_game(dt)

    @property
    def name(self) -> str:
        return "Dolphin"

    @property
    def current_status(self) -> GameConnectionStatus:
        if not self.dolphin.is_hooked():
            return GameConnectionStatus.Disconnected

        if self.patches is None:
            return GameConnectionStatus.UnknownGame

        if self._world is None:
            return GameConnectionStatus.TitleScreen
        elif not self.checking_for_collected_index:
            return GameConnectionStatus.TrackerOnly
        else:
            return GameConnectionStatus.InGame



from random import Random
from typing import Iterator, List, TypeVar, Dict, Any

T = TypeVar('T')


def shuffle(rng: Random, x: Iterator[T]) -> List[T]:
    """
    Shuffles a copy of the given iterator.
    :param rng:
    :param x:
    :return:
    """
    result = list(x)
    rng.shuffle(result)
    return result


def iterate_with_weights(items: Iterator[T],
                         item_weights: Dict[T, float],
                         rng: Random,
                         ) -> Iterator[T]:
    """
    Iterates over the given list randomly, with each item having the probability listed in item_weigths
    :param items:
    :param item_weights:
    :param rng:
    :return:
    """

    items = list(items)
    weights = [max(item_weights[action], 0) for action in items]

    while items and any(weight > 0 for weight in weights):
        pickup_node = rng.choices(items, weights)[0]

        # Remove the pickup_node from the potential list, along with it's weight
        index = items.index(pickup_node)
        items.pop(index)
        weights.pop(index)

        yield pickup_node


def select_element_with_weight(weighted_items: Dict[T, float], rng: Random) -> T:
    return next(iterate_with_weights(items=list(weighted_items.keys()),
                                     item_weights=weighted_items,
                                     rng=rng))


def random_key(d: Dict[T, Any], rng: Random) -> T:
    return rng.choice(list(d.keys()))



from argparse import ArgumentParser
from pathlib import Path

from randovania.games.prime import claris_randomizer
from randovania.generator import generator
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink


def randomize_command_logic(args):
    def status_update(s):
        if args.verbose:
            print(s)

    if args.permalink is not None:
        layout_description = generator.generate_description(permalink=Permalink.from_str(args.permalink),
                                                            status_update=status_update,
                                                            validate_after_generation=True)
    else:
        layout_description = LayoutDescription.from_file(args.log_file)

    cosmetic_patches = CosmeticPatches(
        disable_hud_popup=args.disable_hud_popup,
        speed_up_credits=args.speed_up_credits)

    claris_randomizer.apply_layout(description=layout_description,
                                   cosmetic_patches=cosmetic_patches,
                                   backup_files_path=args.backup_files,
                                   progress_update=lambda x, _: status_update(x),
                                   game_root=args.game_files,
                                   players_config=PlayersConfiguration(0, {0: "You"})
                                   )


def add_randomize_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "randomize",
        help="Randomizes a game files path."
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument("--permalink", type=str, help="The permalink to use")
    group.add_argument("--log-file", type=Path, help="A seed log file to use")

    parser.add_argument("--disable-hud-popup", action="store_true", help="Remove the HUD popup", default=False)
    parser.add_argument("--speed-up-credits", action="store_true", help="Speeds ups the credits sequence",
                        default=False)
    parser.add_argument("--verbose", action="store_true", help="Prints progress",
                        default=False)

    parser.add_argument(
        "--backup-files",
        type=Path,
        help="Folder where to place/restore backups, in case of menu mod.")

    parser.add_argument(
        "game_files",
        type=Path,
        help="Root of an extracted games file to randomize.")
    parser.set_defaults(func=randomize_command_logic)



from random import Random
from unittest.mock import MagicMock

import pytest

from randovania.game_description import data_reader
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.generator.filler import retcon
from randovania.generator.filler.retcon import FillerConfiguration
from randovania.layout.available_locations import RandomizationMode
from randovania.resolver.bootstrap import logic_bootstrap


@pytest.mark.parametrize("major_mode", [RandomizationMode.FULL, RandomizationMode.MAJOR_MINOR_SPLIT])
@pytest.mark.parametrize("has_exclusion", [False, True])
def test_build_available_indices(major_mode: RandomizationMode, has_exclusion: bool):
    # Setup
    world_a = MagicMock()
    world_a.pickup_indices = [PickupIndex(1), PickupIndex(2)]
    world_a.major_pickup_indices = [PickupIndex(1)]

    world_b = MagicMock()
    world_b.pickup_indices = [PickupIndex(3), PickupIndex(4)]
    world_b.major_pickup_indices = [PickupIndex(3)]

    world_list = MagicMock()
    world_list.worlds = [world_a, world_b]

    if has_exclusion:
        exclusion = frozenset([PickupIndex(3)])
    else:
        exclusion = frozenset()
    configuration = FillerConfiguration(major_mode, 0, 0, exclusion)

    # Run
    indices_per_world, all_indices = retcon.build_available_indices(world_list, configuration)

    # Assert
    if major_mode == RandomizationMode.FULL:
        a_pickups = {PickupIndex(1), PickupIndex(2)}
        b_pickups = {PickupIndex(3), PickupIndex(4)}
    else:
        a_pickups = {PickupIndex(1)}
        b_pickups = {PickupIndex(3)}

    if has_exclusion:
        b_pickups.remove(PickupIndex(3))

    assert indices_per_world == [a_pickups, b_pickups]
    assert all_indices == a_pickups | b_pickups


@pytest.mark.skip
@pytest.mark.skip_generation_tests
def test_retcon_filler_integration(default_layout_configuration):
    layout_configuration = default_layout_configuration

    rng = Random("fixed-seed!")
    status_update = MagicMock()

    game = data_reader.decode_data(layout_configuration.game_data)
    patches = game.create_game_patches()
    available_pickups = game.pickup_database.all_useful_pickups

    new_game, state = logic_bootstrap(layout_configuration, game, patches)
    new_game.patch_requirements(state.resources, layout_configuration.damage_strictness.value)

    filler_patches = retcon.retcon_playthrough_filler(new_game,
                                                      state, tuple(available_pickups), rng,
                                                      FillerConfiguration(
                                                          randomization_mode=RandomizationMode.FULL,
                                                          minimum_random_starting_items=0,
                                                          maximum_random_starting_items=0,
                                                          indices_to_exclude=frozenset(),
                                                      ),
                                                      status_update)
    assert filler_patches == patches



import itertools
import json
from pathlib import Path
from typing import List
from unittest.mock import patch, MagicMock, call

import pytest

import randovania.interface_common.options
import randovania.interface_common.persisted_options
from randovania.interface_common import update_checker
from randovania.interface_common.options import Options, DecodeFailedException, InfoAlert


@pytest.fixture(name="option")
def _option() -> Options:
    return Options(MagicMock())


def test_migrate_from_v11(option):
    old_data = {"version": 11,
                "options": {
                    "cosmetic_patches": {
                        "disable_hud_popup": True,
                        "speed_up_credits": True,
                        "open_map": True,
                        "pickup_markers": True,
                    }}}

    # Run
    new_data = randovania.interface_common.persisted_options.get_persisted_options_from_data(old_data)
    option.load_from_persisted(new_data, False)

    # Assert
    expected_data = {
        "cosmetic_patches": {
            "disable_hud_popup": True,
            "speed_up_credits": True,
            "open_map": True,
            "pickup_markers": True,
        }
    }
    assert new_data == expected_data


@patch("randovania.interface_common.options.Options._save_to_disk", autospec=True)
def test_context_manager_with_no_changes_doesnt_save(mock_save_to_disk: MagicMock,
                                                     option: Options):
    # Setup
    settings_changed = MagicMock()
    option.on_options_changed = settings_changed

    # Run
    with option:
        pass

    # Assert
    mock_save_to_disk.assert_not_called()
    settings_changed.assert_not_called()


@patch("randovania.interface_common.options.Options._save_to_disk", autospec=True)
def test_save_with_context_manager(mock_save_to_disk: MagicMock,
                                   option: Options):
    # Setup
    settings_changed = MagicMock()
    option._output_directory = Path("start")
    option.on_options_changed = settings_changed

    # Run
    with option:
        option.output_directory = Path("end")

    # Assert
    mock_save_to_disk.assert_called_once_with(option)
    settings_changed.assert_called_once_with()


@patch("randovania.interface_common.options.Options._save_to_disk", autospec=True)
def test_single_save_with_nested_context_manager(mock_save_to_disk: MagicMock,
                                                 option: Options):
    # Setup
    option._output_directory = Path("start")

    # Run
    with option:
        option.output_directory = Path("end")
        with option:
            pass

    # Assert
    mock_save_to_disk.assert_called_once_with(option)


def test_changing_field_without_context_manager_should_error(option: Options):
    # Run
    with pytest.raises(AssertionError) as exception:
        option.output_directory = Path("start")

    # Assert
    assert str(exception.value) == "Attempting to edit an Options, but it wasn't made editable"


@patch("randovania.interface_common.options.get_persisted_options_from_data", autospec=True)
@patch("randovania.interface_common.options.Options._read_persisted_options", return_value=None, autospec=True)
def test_load_from_disk_no_data(mock_read_persisted_options: MagicMock,
                                mock_get_persisted_options_from_data: MagicMock,
                                option: Options):
    # Run
    option.load_from_disk()

    # Assert
    mock_read_persisted_options.assert_called_once_with(option)
    mock_get_persisted_options_from_data.assert_not_called()


@pytest.mark.parametrize("fields_to_test",
                         itertools.combinations(randovania.interface_common.options._SERIALIZER_FOR_FIELD.keys(), 2))
@patch("randovania.interface_common.options.Options._set_field", autospec=True)
@patch("randovania.interface_common.options.get_persisted_options_from_data", autospec=True)
@patch("randovania.interface_common.options.Options._read_persisted_options", autospec=True)
def test_load_from_disk_with_data(mock_read_persisted_options: MagicMock,
                                  mock_get_persisted_options_from_data: MagicMock,
                                  mock_set_field: MagicMock,
                                  fields_to_test: List[str],
                                  option: Options):
    # Setup
    persisted_options = {
        field_name: MagicMock()
        for field_name in fields_to_test
    }
    new_serializers = {
        field_name: MagicMock()
        for field_name in fields_to_test
    }
    mock_get_persisted_options_from_data.return_value = persisted_options

    # Run
    with patch.dict(randovania.interface_common.options._SERIALIZER_FOR_FIELD, new_serializers):
        option.load_from_disk()

    # Assert
    mock_read_persisted_options.assert_called_once_with(option)
    mock_get_persisted_options_from_data.assert_called_once_with(mock_read_persisted_options.return_value)
    for field_name, serializer in new_serializers.items():
        serializer.decode.assert_called_once_with(persisted_options[field_name])

    mock_set_field.assert_has_calls(
        call(option, field_name, new_serializers[field_name].decode.return_value)
        for field_name in fields_to_test
    )


# TODO: test with an actual field as well
def test_serialize_fields(option: Options):
    # Setup

    # Run
    result = option._serialize_fields()

    # Assert
    assert result == {
        "version": randovania.interface_common.persisted_options._CURRENT_OPTIONS_FILE_VERSION,
        "options": {
            "last_changelog_displayed": str(update_checker.strict_current_version()),
        }
    }


def test_edit_during_options_changed(tmpdir):
    # Setup
    option = Options(Path(tmpdir))
    option._output_directory = Path("start")

    def on_changed():
        with option:
            option.output_directory = Path("final")

    option.on_options_changed = on_changed

    # Run
    with option:
        option.output_directory = Path("middle")

    second_option = Options(Path(tmpdir))
    second_option.load_from_disk()

    # Assert
    assert option.output_directory == Path("final")
    assert option.output_directory == second_option.output_directory


@pytest.mark.parametrize("ignore_decode_errors", [False, True])
def test_load_from_disk_missing_json(ignore_decode_errors: bool,
                                     tmpdir):
    # Setup
    option = Options(Path(tmpdir))
    tmpdir.join("config.json").write_text("", "utf-8")

    if ignore_decode_errors:
        result = option.load_from_disk(ignore_decode_errors)
        assert result != ignore_decode_errors
    else:
        with pytest.raises(DecodeFailedException):
            option.load_from_disk(ignore_decode_errors)


@pytest.mark.parametrize("ignore_decode_errors", [False])
def test_load_from_disk_invalid_json(ignore_decode_errors: bool,
                                     tmpdir):
    # Setup
    option = Options(Path(tmpdir))
    tmpdir.join("config.json").write_text(
        json.dumps(randovania.interface_common.persisted_options.serialized_data_for_options({
            "cosmetic_patches": {
                "pickup_model_style": "invalid-value"
            }
        })),
        "utf-8")

    if ignore_decode_errors:
        result = option.load_from_disk(ignore_decode_errors)
        assert result != ignore_decode_errors
    else:
        with pytest.raises(DecodeFailedException):
            option.load_from_disk(ignore_decode_errors)


def test_reset_to_defaults():
    # Create and test they're equal
    blank = Options(MagicMock())
    modified = Options(MagicMock())
    assert blank._serialize_fields() == modified._serialize_fields()

    # Modify and test they're different
    with modified:
        for field in randovania.interface_common.options._SERIALIZER_FOR_FIELD.keys():
            # This cause weirdness in serializing it
            if field != "last_changelog_displayed":
                modified._set_field(field, getattr(modified, field))
    assert blank._serialize_fields() != modified._serialize_fields()

    # Reset and test they're the same
    with modified:
        modified.reset_to_defaults()
    assert blank._serialize_fields() == modified._serialize_fields()


def test_setting_fields_to_self_do_nothing():
    options = Options(MagicMock())
    initial_serialize = options._serialize_fields()

    # Modify and test they're different
    with options:
        for field in randovania.interface_common.options._SERIALIZER_FOR_FIELD.keys():
            setattr(options, field, getattr(options, field))

    # Reset and test they're the same
    assert options._serialize_fields() == initial_serialize


def test_mark_alert_as_displayed(tmp_path):
    opt = Options(tmp_path)

    all_alerts = (InfoAlert.FAQ, InfoAlert.MULTIWORLD_FAQ)
    for alert in all_alerts:
        assert not opt.is_alert_displayed(alert)
        opt.mark_alert_as_displayed(alert)
        assert opt.is_alert_displayed(alert)

    assert opt.displayed_alerts == set(all_alerts)

    new_opt = Options(tmp_path)
    new_opt.load_from_disk()
    assert new_opt.displayed_alerts == set(all_alerts)



import copy
import dataclasses
from enum import Enum


def _int_field(default: int, min_value: int, max_value: int, display_as_percentage: bool = True):
    return dataclasses.field(default=default, metadata={"min": min_value, "max": max_value,
                                                        "display_as_percentage": display_as_percentage})


class SoundMode(Enum):
    MONO = 0
    STEREO = 1
    SURROUND = 2


@dataclasses.dataclass(frozen=True)
class EchoesUserPreferences:
    sound_mode: SoundMode = SoundMode.STEREO
    screen_brightness: int = _int_field(4, 0, 8)
    screen_x_offset: int = _int_field(0, -0x1e, 0x1f, False)
    screen_y_offset: int = _int_field(0, -0x1e, 0x1f, False)
    screen_stretch: int = _int_field(0, -10, 10, False)
    sfx_volume: int = _int_field(0x69, 0x00, 0x69)
    music_volume: int = _int_field(0x4f, 0x00, 0x69)
    hud_alpha: int = _int_field(0xff, 0x00, 0xff)
    helmet_alpha: int = _int_field(0xff, 0x00, 0xff)
    hud_lag: bool = True
    invert_y_axis: bool = False
    rumble: bool = True
    hint_system: bool = False

    def __post_init__(self):
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if "min" in field.metadata and value < field.metadata["min"]:
                raise ValueError(f'Value {value} for field "{field.name}" is less than minimum {field.metadata["min"]}')
            if "max" in field.metadata and value > field.metadata["max"]:
                raise ValueError(f'Value {value} for field "{field.name}" is less than maximum {field.metadata["max"]}')

    @property
    def as_json(self) -> dict:
        result = {
            field.name: getattr(self, field.name)
            for field in dataclasses.fields(self)
        }
        result["sound_mode"] = self.sound_mode.value
        return result

    @classmethod
    def from_json_dict(cls, json_dict: dict) -> "EchoesUserPreferences":
        kwargs = copy.copy(json_dict)
        kwargs["sound_mode"] = SoundMode(kwargs["sound_mode"])
        return cls(**kwargs)



import re

import discord

import randovania
from randovania.games.game import RandovaniaGame
from randovania.gui.lib import preset_describer
from randovania.layout.permalink import Permalink

_PRETTY_GAME_NAME = {
    RandovaniaGame.PRIME1: "Metroid Prime 1",
    RandovaniaGame.PRIME2: "Metroid Prime 2: Echoes",
    RandovaniaGame.PRIME3: "Metroid Prime 3: Corruption",
}
possible_links_re = re.compile(r'([A-Za-z0-9-_]{8,})')


async def look_for_permalinks(message: str, channel: discord.TextChannel):
    for word in possible_links_re.finditer(message):
        try:
            permalink = Permalink.from_str(word.group(1))
        except ValueError:
            continue

        embed = discord.Embed(title=word.group(1),
                              description="{} player multiworld permalink".format(permalink.player_count))

        if permalink.player_count == 1:
            preset = permalink.get_preset(0)
            embed.description = "{} permalink for Randovania {}".format(_PRETTY_GAME_NAME[preset.game],
                                                                        randovania.VERSION)
            for category, items in preset_describer.describe(preset):
                embed.add_field(name=category, value="\n".join(items), inline=True)

        await channel.send(embed=embed)


class Bot(discord.Client):
    def __init__(self, configuration: dict):
        super().__init__()
        self.configuration = configuration

    async def on_message(self, message: discord.Message):
        if message.author == self.user:
            return

        channel: discord.TextChannel = message.channel
        if self.configuration["channel_name_filter"] not in channel.name:
            return

        if message.guild.id != self.configuration["guild"]:
            return

        await look_for_permalinks(message.content, channel)


def run():
    configuration = randovania.get_configuration()["discord_bot"]

    client = Bot(configuration)
    client.run(configuration["token"])



import dataclasses
import multiprocessing.dummy
from random import Random
from typing import Iterator, Optional, Callable, List, Dict

import tenacity

from randovania import VERSION
from randovania.game_description import data_reader, default_database
from randovania.game_description.assignment import PickupAssignment, PickupTarget
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.world_list import WorldList
from randovania.generator import base_patches_factory
from randovania.generator.filler.filler_library import filter_unassigned_pickup_nodes, UnableToGenerate
from randovania.generator.filler.runner import run_filler, FillerPlayerResult, PlayerPool, FillerResults
from randovania.generator.item_pool import pool_creator
from randovania.layout.available_locations import RandomizationMode
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.layout.preset import Preset
from randovania.resolver import resolver
from randovania.resolver.exceptions import GenerationFailure, InvalidConfiguration, ImpossibleForSolver
from randovania.resolver.state import State


def _iterate_previous_states(state: State) -> Iterator[State]:
    while state:
        yield state
        state = state.previous_state


def generate_description(permalink: Permalink,
                         status_update: Optional[Callable[[str], None]],
                         validate_after_generation: bool,
                         timeout: Optional[int] = 600,
                         attempts: int = 15,
                         ) -> LayoutDescription:
    """
    Creates a LayoutDescription for the given Permalink.
    :param permalink:
    :param status_update:
    :param validate_after_generation:
    :param timeout: Abort generation after this many seconds.
    :param attempts: Attempt this many generations.
    :return:
    """
    if status_update is None:
        status_update = id

    try:
        result = _async_create_description(
            permalink=permalink,
            status_update=status_update,
            attempts=attempts,
        )
    except UnableToGenerate as e:
        raise GenerationFailure("Could not generate a game with the given settings",
                                permalink=permalink, source=e) from e

    if validate_after_generation and permalink.player_count == 1:
        with multiprocessing.dummy.Pool(1) as dummy_pool:
            resolve_params = {
                "configuration": permalink.presets[0].configuration,
                "patches": result.all_patches[0],
                "status_update": status_update,
            }
            final_state_async = dummy_pool.apply_async(func=resolver.resolve,
                                                       kwds=resolve_params)

            try:
                final_state_by_resolve = final_state_async.get(timeout)
            except multiprocessing.TimeoutError as e:
                raise GenerationFailure("Timeout reached when validating possibility",
                                        permalink=permalink, source=e) from e

            if final_state_by_resolve is None:
                # Why is final_state_by_distribution not OK?
                raise GenerationFailure("Generated game was considered impossible by the solver",
                                        permalink=permalink, source=ImpossibleForSolver())

    return result


def _validate_item_pool_size(item_pool: List[PickupEntry], game: GameDescription,
                             configuration: EchoesConfiguration) -> None:
    min_starting_items = configuration.major_items_configuration.minimum_random_starting_items
    if len(item_pool) > game.world_list.num_pickup_nodes + min_starting_items:
        raise InvalidConfiguration(
            "Item pool has {} items, which is more than {} (game) + {} (minimum starting items)".format(
                len(item_pool), game.world_list.num_pickup_nodes, min_starting_items))


def _distribute_remaining_items(rng: Random,
                                filler_results: Dict[int, FillerPlayerResult],
                                ) -> Dict[int, GamePatches]:
    unassigned_pickup_nodes = []
    all_remaining_pickups = []
    assignments: Dict[int, PickupAssignment] = {}

    for index, filler_result in filler_results.items():
        for pickup_node in filter_unassigned_pickup_nodes(filler_result.game.world_list.all_nodes,
                                                          filler_result.patches.pickup_assignment):
            unassigned_pickup_nodes.append((index, pickup_node))

        all_remaining_pickups.extend(zip([index] * len(filler_result.unassigned_pickups),
                                         filler_result.unassigned_pickups))
        assignments[index] = {}

    rng.shuffle(unassigned_pickup_nodes)
    rng.shuffle(all_remaining_pickups)

    for (node_player, node), (pickup_player, pickup) in zip(unassigned_pickup_nodes, all_remaining_pickups):
        assignments[node_player][node.pickup_index] = PickupTarget(pickup, pickup_player)

    return {
        index: filler_results[index].patches.assign_pickup_assignment(assignment)
        for index, assignment in assignments.items()
    }
    # FIXME: ignoring major-minor randomization

    # return {
    #     index: patches.assign_pickup_assignment(
    #         _assign_remaining_items(rng, game.world_list, patches.pickup_assignment, remaining_items,
    #                                 configuration.randomization_mode)
    #     )
    #     for index, (patches, remaining_items) in _retryable_create_patches(rng, game, presets, status_update).items()
    # }


def _async_create_description(permalink: Permalink,
                              status_update: Callable[[str], None],
                              attempts: int,
                              ) -> LayoutDescription:
    """
    :param permalink:
    :param status_update:
    :return:
    """
    rng = Random(permalink.as_bytes)

    presets = {
        i: permalink.get_preset(i)
        for i in range(permalink.player_count)
    }

    retrying = tenacity.Retrying(
        stop=tenacity.stop_after_attempt(attempts),
        retry=tenacity.retry_if_exception_type(UnableToGenerate),
        reraise=True
    )

    filler_results = retrying(_create_pools_and_fill, rng, presets, status_update)
    all_patches = _distribute_remaining_items(rng, filler_results.player_results)
    return LayoutDescription(
        permalink=permalink,
        version=VERSION,
        all_patches=all_patches,
        item_order=filler_results.action_log,
    )


def create_player_pool(rng: Random, configuration: EchoesConfiguration,
                       player_index: int, num_players: int) -> PlayerPool:
    game = default_database.game_description_for(configuration.game)
    base_patches = dataclasses.replace(base_patches_factory.create_base_patches(configuration, rng, game,
                                                                                num_players > 1),
                                       player_index=player_index)

    item_pool, pickup_assignment, initial_items = pool_creator.calculate_pool_results(configuration,
                                                                                      game.resource_database)
    target_assignment = {
        index: PickupTarget(pickup, player_index)
        for index, pickup in pickup_assignment.items()
    }
    patches = base_patches.assign_pickup_assignment(target_assignment).assign_extra_initial_items(initial_items)

    return PlayerPool(
        game=game,
        configuration=configuration,
        patches=patches,
        pickups=item_pool,
    )


def _create_pools_and_fill(rng: Random,
                           presets: Dict[int, Preset],
                           status_update: Callable[[str], None],
                           ) -> FillerResults:
    """
    Runs the rng-dependant parts of the generation, with retries
    :param rng:
    :param presets:
    :param status_update:
    :return:
    """
    player_pools: Dict[int, PlayerPool] = {}

    for player_index, player_preset in presets.items():
        status_update(f"Creating item pool for player {player_index + 1}")
        player_pools[player_index] = create_player_pool(rng, player_preset.configuration, player_index,
                                                        len(presets))

    for player_pool in player_pools.values():
        _validate_item_pool_size(player_pool.pickups, player_pool.game, player_pool.configuration)

    return run_filler(rng, player_pools, status_update)


def _assign_remaining_items(rng: Random,
                            world_list: WorldList,
                            pickup_assignment: PickupAssignment,
                            remaining_items: List[PickupEntry],
                            randomization_mode: RandomizationMode,
                            ) -> PickupAssignment:
    """

    :param rng:
    :param world_list:
    :param pickup_assignment:
    :param remaining_items:
    :return:
    """

    unassigned_pickup_nodes = list(filter_unassigned_pickup_nodes(world_list.all_nodes, pickup_assignment))

    num_etm = len(unassigned_pickup_nodes) - len(remaining_items)
    if num_etm < 0:
        raise InvalidConfiguration(
            "Received {} remaining items, but there's only {} unassigned pickups".format(len(remaining_items),
                                                                                         len(unassigned_pickup_nodes)))

    # Shuffle the items to add and the spots to choose from
    rng.shuffle(remaining_items)
    rng.shuffle(unassigned_pickup_nodes)

    assignment = {}

    if randomization_mode is RandomizationMode.MAJOR_MINOR_SPLIT:
        remaining_majors = [item for item in remaining_items if not item.is_expansion] + ([None] * num_etm)
        unassigned_major_locations = [pickup_node for pickup_node in unassigned_pickup_nodes if
                                      pickup_node.major_location]

        for pickup_node, item in zip(unassigned_major_locations, remaining_majors):
            if item is not None:
                assignment[pickup_node.pickup_index] = item
                remaining_items.remove(item)
            unassigned_pickup_nodes.remove(pickup_node)

    assignment.update({
        pickup_node.pickup_index: item
        for pickup_node, item in zip(unassigned_pickup_nodes, remaining_items)
    })
    return assignment



from typing import Dict, Optional

import networkx

from randovania.game_description.area import Area
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import Node, DockNode, TeleporterNode, PickupNode, ResourceNode
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.world_list import WorldList


def distances_to_node(world_list: WorldList, starting_node: Node,
                      *,
                      ignore_elevators: bool = True,
                      cutoff: Optional[int] = None,
                      patches: Optional[GamePatches] = None,
                      ) -> Dict[Area, int]:
    """
    Compute the shortest distance from a node to all reachable areas.
    :param world_list:
    :param starting_node:
    :param ignore_elevators:
    :param cutoff: Exclude areas with a length longer that cutoff.
    :param patches:
    :return: Dict keyed by area to shortest distance to starting_node.
    """
    g = networkx.DiGraph()

    dock_connections = patches.dock_connection if patches is not None else {}
    elevator_connections = patches.elevator_connection if patches is not None else {}

    for area in world_list.all_areas:
        g.add_node(area)

    for world in world_list.worlds:
        for area in world.areas:
            new_areas = set()
            for node in area.nodes:
                if isinstance(node, DockNode):
                    connection = dock_connections.get((area.area_asset_id, node.dock_index), node.default_connection)
                    new_areas.add(world.area_by_asset_id(connection.area_asset_id))
                elif isinstance(node, TeleporterNode) and not ignore_elevators:
                    connection = elevator_connections.get(node.teleporter_instance_id, node.default_connection)
                    new_areas.add(world_list.area_by_area_location(connection))

            for next_area in new_areas:
                g.add_edge(area, next_area)

    return networkx.single_source_shortest_path_length(g, world_list.nodes_to_area(starting_node), cutoff)


def pickup_index_to_node(world_list: WorldList, index: PickupIndex) -> PickupNode:
    for node in world_list.all_nodes:
        if isinstance(node, PickupNode) and node.pickup_index == index:
            return node
    raise ValueError(f"PickupNode with {index} not found.")


def node_with_resource(world_list: WorldList, resource: ResourceInfo) -> ResourceNode:
    for node in world_list.all_nodes:
        if isinstance(node, ResourceNode) and node.resource() == resource:
            return node
    raise ValueError(f"ResourceNode with {resource} not found.")



import asyncio
import struct

next_client_id = 0


def create_client_id():
    global next_client_id
    next_client_id += 1
    return next_client_id


_hard_coded_reads = {
    # Is this Prime 2 NTSC? Yes.
    b'\x00\x01\x01\x01\x80:\xc3\xb0\x806': b"\x01!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32",
    b'\x00\x01\x01\x01\x80A\x8e\xb8\xb0\x00\x04': b"\x01\x3B\xFA\x3E\xFF",
}


async def client_handler(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
    client_id = create_client_id()

    print(f"[{client_id: 3}] Client connected")
    while True:
        command = await reader.read(1024)
        command_id, num_ops, num_addresses, keep_alive = struct.unpack_from(">BBBB", command)
        print(f"[{client_id: 3}] Received request", command_id, num_ops, num_addresses, keep_alive)

        await asyncio.sleep(1)
        if command_id == 1:
            writer.write(struct.pack(">IIII", 2, 100, 100, 4))
        elif command_id == 0:
            if command in _hard_coded_reads:
                writer.write(_hard_coded_reads[command])
            else:
                print(f"[{client_id: 3}] Received command: {command}")
                writer.write(b"\x00\x00")

        await writer.drain()
        if not bool(keep_alive):
            writer.close()
            break


async def main():
    print("start_server")
    server = await asyncio.start_server(client_handler, port=43673)

    async with server:
        await server.serve_forever()


if __name__ == '__main__':
    asyncio.run(main())



import dataclasses
from enum import Enum
from typing import List

from randovania.bitpacking.bitpacking import BitPackEnum, BitPackDataClass
from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.games.game import RandovaniaGame
from randovania.games.prime import default_data
from randovania.layout.base_configuration import BaseConfiguration
from randovania.layout.beam_configuration import BeamConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.hint_configuration import HintConfiguration
from randovania.layout.translator_configuration import TranslatorConfiguration


@dataclasses.dataclass(frozen=True)
class PrimeConfiguration(BaseConfiguration):
    elevators: LayoutElevators
    skip_final_bosses: bool
    energy_per_tank: float = dataclasses.field(metadata={"min": 1.0, "max": 1000.0, "precision": 1.0})
    dangerous_energy_tank: bool

    @classmethod
    def game_enum(cls) -> RandovaniaGame:
        return RandovaniaGame.PRIME1

    def dangerous_settings(self) -> List[str]:
        result = super().dangerous_settings()
        
        if self.elevators == LayoutElevators.ONE_WAY_ANYTHING:
            result.append("One-way anywhere elevators")

        return result



import copy
import json

import pytest

from randovania import get_data_path
from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description import default_database
from randovania.games.game import RandovaniaGame
from randovania.layout.ammo_configuration import AmmoConfiguration


@pytest.fixture(
    params=[
        {"game": RandovaniaGame.PRIME2, "encoded": b'\x00\x00', "items_state": {}},
        {"game": RandovaniaGame.PRIME2, "encoded": b'"\x80\x00', "maximum_ammo": {"45": 20}},
        {"game": RandovaniaGame.PRIME2, "encoded": b'\t\x90',
         "items_state": {"Missile Expansion": {"variance": 0, "pickup_count": 12}}},
        {"game": RandovaniaGame.PRIME2, "encoded": b'"\x89\x90', "maximum_ammo": {"45": 20},
         "items_state": {"Missile Expansion": {"variance": 0, "pickup_count": 12}}},
        {"game": RandovaniaGame.PRIME3, "encoded": b'\x00', "items_state": {}},
        {"game": RandovaniaGame.PRIME3, "encoded": b'\x99\x00', "maximum_ammo": {"4": 50}},
    ],
    name="config_with_data")
def _config_with_data(request):
    game: RandovaniaGame = request.param["game"]

    with get_data_path().joinpath("item_database", game.value, "default_state", "ammo.json").open() as open_file:
        default_data = json.load(open_file)

    default = AmmoConfiguration.from_json(default_data, default_database.item_database_for_game(game))
    data = copy.deepcopy(default_data)

    for key, value in request.param.get("items_state", {}).items():
        data["items_state"][key] = value

    for key, value in request.param.get("maximum_ammo", {}).items():
        data["maximum_ammo"][key] = value

    config = AmmoConfiguration.from_json(data, default_database.item_database_for_game(game))
    return request.param["encoded"], config, default


def test_decode(config_with_data):
    # Setup
    data, expected, reference = config_with_data

    # Run
    decoder = BitPackDecoder(data)
    result = AmmoConfiguration.bit_pack_unpack(decoder, {"reference": reference})

    # Assert
    assert result == expected


def test_encode(config_with_data):
    # Setup
    expected, value, reference = config_with_data

    # Run
    result = bitpacking.pack_value(value, metadata={"reference": reference})

    # Assert
    assert result == expected



import datetime
import functools
import json
from typing import Iterator, List, Optional, Callable, Any

import peewee

from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.preset import Preset
from randovania.layout.preset_migration import VersionedPreset
from randovania.network_common.session_state import GameSessionState

db = peewee.SqliteDatabase(None, pragmas={'foreign_keys': 1})


class BaseModel(peewee.Model):
    class Meta:
        database = db


class EnumField(peewee.CharField):
    """
    This class enable an Enum like field for Peewee
    """

    def __init__(self, choices: Callable, *args: Any, **kwargs: Any) -> None:
        super(peewee.CharField, self).__init__(*args, **kwargs)
        self.choices = choices
        self.max_length = 255

    def db_value(self, value: Any) -> Any:
        return value.value

    def python_value(self, value: Any) -> Any:
        return self.choices(type(list(self.choices)[0].value)(value))


class User(BaseModel):
    discord_id = peewee.IntegerField(index=True, null=True)
    name = peewee.CharField()

    @classmethod
    def get_by_id(cls, pk) -> "User":
        return cls.get(cls._meta.primary_key == pk)

    @property
    def as_json(self):
        return {
            "id": self.id,
            "name": self.name,
        }


@functools.lru_cache()
def _decode_layout_description(s):
    return LayoutDescription.from_json_dict(json.loads(s))


def _datetime_now():
    return datetime.datetime.now(datetime.timezone.utc)


class GameSession(BaseModel):
    name = peewee.CharField()
    password = peewee.CharField(null=True)
    state = EnumField(choices=GameSessionState, default=GameSessionState.SETUP)
    layout_description_json = peewee.TextField(null=True)
    seed_hash = peewee.CharField(null=True)
    creator = peewee.ForeignKeyField(User)
    creation_date = peewee.DateTimeField(default=_datetime_now)
    generation_in_progress = peewee.ForeignKeyField(User, null=True)

    @property
    def all_presets(self) -> List[Preset]:
        return [
            VersionedPreset(json.loads(preset.preset)).get_preset()
            for preset in sorted(self.presets, key=lambda it: it.row)
        ]

    @property
    def num_rows(self) -> int:
        return len(self.presets)

    @property
    def layout_description(self) -> Optional[LayoutDescription]:
        # FIXME: a server can have an invalid layout description. Likely from an old version!
        return _decode_layout_description(self.layout_description_json) if self.layout_description_json else None

    @layout_description.setter
    def layout_description(self, description: Optional[LayoutDescription]):
        self.layout_description_json = json.dumps(description.as_json) if description is not None else None

    @property
    def creation_datetime(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self.creation_date)

    def create_list_entry(self):
        return {
            "id": self.id,
            "name": self.name,
            "has_password": self.password is not None,
            "state": self.state.value,
            "num_players": len(self.players),
            "creator": self.creator.name,
            "creation_date": self.creation_datetime.astimezone(datetime.timezone.utc).isoformat(),
        }

    def create_session_entry(self):
        description = self.layout_description

        location_to_name = {
            row: f"Player {row + 1}" for row in range(self.num_rows)
        }
        for membership in self.players:
            if not membership.is_observer:
                location_to_name[membership.row] = membership.effective_name

        def _describe_action(action: GameSessionTeamAction) -> dict:
            provider: int = action.provider_row
            receiver: int = action.receiver_row
            time = datetime.datetime.fromisoformat(action.time)
            target = description.all_patches[provider].pickup_assignment[PickupIndex(action.provider_location_index)]

            message = (f"{location_to_name[provider]} found {target.pickup.name} "
                       f"for {location_to_name[receiver]}.")

            return {
                "message": message,
                "time": time.astimezone(datetime.timezone.utc).isoformat(),
            }

        if description is not None:
            game_details = {
                "spoiler": description.permalink.spoiler,
                "word_hash": description.shareable_word_hash,
                "seed_hash": description.shareable_hash,
                "permalink": description.permalink.as_base64_str,
            }
        else:
            game_details = {
                "spoiler": None,
                "word_hash": None,
                "seed_hash": None,
                "permalink": None,
            }

        return {
            "id": self.id,
            "name": self.name,
            "state": self.state.value,
            "players": [
                membership.as_json
                for membership in self.players
            ],
            "presets": [
                json.loads(preset.preset)
                for preset in sorted(self.presets, key=lambda it: it.row)
            ],
            "actions": [
                _describe_action(action)
                for action in GameSessionTeamAction.select().where(GameSessionTeamAction.session == self
                                                                   ).order_by(GameSessionTeamAction.time.asc())
            ],
            **game_details,
            "generation_in_progress": (self.generation_in_progress.id
                                       if self.generation_in_progress is not None else None),
        }

    def reset_layout_description(self):
        self.layout_description_json = None
        self.save()


class GameSessionPreset(BaseModel):
    session = peewee.ForeignKeyField(GameSession, backref="presets")
    row = peewee.IntegerField()
    preset = peewee.TextField()

    class Meta:
        primary_key = peewee.CompositeKey('session', 'row')


class GameSessionMembership(BaseModel):
    user = peewee.ForeignKeyField(User, backref="games")
    session = peewee.ForeignKeyField(GameSession, backref="players")
    row = peewee.IntegerField(null=True)
    admin = peewee.BooleanField()
    join_date = peewee.DateTimeField(default=_datetime_now)
    connection_state = peewee.TextField(null=True)
    inventory = peewee.TextField(null=True)

    @property
    def as_json(self):
        return {
            "id": self.user.id,
            "name": self.user.name,
            "row": self.row,
            "admin": self.admin,
            "inventory": self.inventory,
            "connection_state": self.connection_state,
        }

    @property
    def effective_name(self) -> str:
        return self.user.name

    @property
    def is_observer(self) -> bool:
        return self.row is None

    @classmethod
    def get_by_ids(cls, user_id: int, session_id: int) -> "GameSessionMembership":
        return GameSessionMembership.get(
            GameSessionMembership.session == session_id,
            GameSessionMembership.user == user_id,
        )

    @classmethod
    def get_by_session_position(cls, session: GameSession, row: int) -> "GameSessionMembership":
        return GameSessionMembership.get(
            GameSessionMembership.session == session,
            GameSessionMembership.row == row,
        )

    @classmethod
    def non_observer_members(cls, session: GameSession) -> Iterator["GameSessionMembership"]:
        yield from GameSessionMembership.select().where(GameSessionMembership.session == session,
                                                        GameSessionMembership.row != None,
                                                        )

    class Meta:
        primary_key = peewee.CompositeKey('user', 'session')
        constraints = [peewee.SQL('UNIQUE(session_id, row)')]


class GameSessionTeamAction(BaseModel):
    session = peewee.ForeignKeyField(GameSession)
    provider_row = peewee.IntegerField()
    provider_location_index = peewee.IntegerField()
    receiver_row = peewee.IntegerField()

    time = peewee.DateTimeField(default=_datetime_now)

    class Meta:
        primary_key = peewee.CompositeKey('session', 'provider_row', 'provider_location_index')


all_classes = [User, GameSession, GameSessionPreset, GameSessionMembership, GameSessionTeamAction]



from unittest.mock import MagicMock, call

import pytest
from mock import AsyncMock

from randovania.game_connection.connection_backend import MemoryOperation
from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_connection.dolphin_backend import DolphinBackend


@pytest.fixture(name="backend")
def dolphin_backend():
    backend = DolphinBackend()
    backend.dolphin = MagicMock()
    return backend


@pytest.mark.parametrize("depth", [0, 1, 2, 3])
@pytest.mark.asyncio
async def test_update(backend, depth: int):
    # Setup
    backend._enabled = depth > 0
    backend._ensure_hooked = MagicMock(return_value=depth < 1)
    backend._identify_game = AsyncMock(return_value=depth > 2)
    backend._interact_with_game = AsyncMock()

    # Run
    await backend.update(1)

    # Assert
    backend._ensure_hooked.assert_has_calls([call()] if depth > 0 else [])
    backend._identify_game.assert_has_awaits([call()] if depth > 1 else [])
    backend._interact_with_game.assert_has_awaits([call(1)] if depth > 2 else [])


def test_current_status_disconnected(backend):
    backend.dolphin.is_hooked.return_value = False
    assert backend.current_status == GameConnectionStatus.Disconnected


def test_current_status_wrong_game(backend):
    backend.dolphin.is_hooked.return_value = True
    assert backend.current_status == GameConnectionStatus.UnknownGame


def test_current_status_not_in_game(backend):
    backend.dolphin.is_hooked.return_value = True
    backend.patches = True
    assert backend.current_status == GameConnectionStatus.TitleScreen


def test_current_status_tracker_only(backend):
    backend.dolphin.is_hooked.return_value = True
    backend.patches = True
    backend._world = True
    assert backend.current_status == GameConnectionStatus.TrackerOnly


def test_current_status_in_game(backend):
    backend.dolphin.is_hooked.return_value = True
    backend.patches = True
    backend._world = True
    backend.checking_for_collected_index = True
    assert backend.current_status == GameConnectionStatus.InGame


@pytest.mark.asyncio
async def test_perform_memory_operations(backend: DolphinBackend):
    backend.dolphin.follow_pointers.return_value = 0x80003000
    backend.dolphin.read_bytes.side_effect = [b"A" * 50, b"B" * 30, b"C" * 10]

    # Run
    result = await backend._perform_memory_operations([
        MemoryOperation(0x80001000, offset=20, read_byte_count=50),
        MemoryOperation(0x80001000, offset=10, read_byte_count=30, write_bytes=b"1" * 30),
        MemoryOperation(0x80002000, read_byte_count=10),
    ])

    # Assert
    assert list(result.values()) == [b"A" * 50, b"B" * 30, b"C" * 10]
    backend.dolphin.follow_pointers.assert_called_once_with(0x80001000, [0x0])
    backend.dolphin.read_bytes.assert_has_calls([
        call(0x80003000 + 20, 50),
        call(0x80003000 + 10, 30),
        call(0x80002000, 10),
    ])
    backend.dolphin.write_bytes.assert_called_once_with(0x80003000 + 10, b"1" * 30)



from randovania.game_description.resources.resource_type import ResourceType


class PickupIndex:
    _index: int

    @property
    def resource_type(self) -> ResourceType:
        return ResourceType.PICKUP_INDEX

    @property
    def long_name(self) -> str:
        return "PickupIndex {}".format(self._index)

    def __init__(self, index: int):
        self._index = index

    def __lt__(self, other: "PickupIndex") -> bool:
        return self._index < other._index

    def __repr__(self):
        return self.long_name

    def __hash__(self):
        return self._index

    def __eq__(self, other):
        return isinstance(other, PickupIndex) and other._index == self._index

    @property
    def index(self) -> int:
        return self._index

    @property
    def as_json(self):
        return self._index

    @classmethod
    def from_json(cls, value):
        return cls(value)



from pathlib import Path
from unittest.mock import patch, MagicMock

from randovania.games.prime import dol_patcher


@patch("randovania.games.prime.dol_patcher.DolFile")
@patch("randovania.games.prime.dol_patcher._get_dol_path", autospec=True)
@patch("randovania.games.prime.dol_patcher.find_version_for_dol", autospec=True)
def test_apply_patches(mock_find_version_for_dol: MagicMock,
                       mock_get_dol_path: MagicMock,
                       mock_dol_file_constructor: MagicMock,
                       mocker,
                       ):
    # Setup
    game_root = MagicMock()
    game_patches = MagicMock()
    user_preferences = MagicMock()
    version_patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    mock_find_version_for_dol.return_value = version_patches
    dol_file = mock_dol_file_constructor.return_value

    mock_apply_string = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.apply_remote_execution_patch", autospec=True)
    mock_apply_game_options: MagicMock = mocker.patch(
        "randovania.games.prime.echoes_dol_patches.apply_game_options_patch", autospec=True)
    mock_apply_capacity: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.apply_energy_tank_capacity_patch", autospec=True)
    mock_apply_beam_cost_patch: MagicMock = mocker.patch(
        "randovania.games.prime.echoes_dol_patches.apply_beam_cost_patch", autospec=True)
    mock_apply_starting_visor_patch: MagicMock = mocker.patch(
        "randovania.games.prime.echoes_dol_patches.apply_starting_visor_patch", autospec=True)

    # Run
    dol_patcher.apply_patches(game_root, game_patches.game_specific, user_preferences, {"foo": "bar"})

    # Assert
    mock_find_version_for_dol.assert_called_once_with(dol_file, dol_patcher.ALL_VERSIONS_PATCHES)
    mock_get_dol_path.assert_called_once_with(game_root)
    mock_dol_file_constructor.assert_called_once_with(mock_get_dol_path.return_value)
    mock_apply_string.assert_called_once_with(version_patches.string_display, dol_file)
    mock_apply_game_options.assert_called_once_with(
        version_patches.game_options_constructor_address,
        user_preferences, dol_file
    )
    mock_apply_capacity.assert_called_once_with(
        version_patches.health_capacity,
        game_patches.game_specific, dol_file
    )
    mock_apply_beam_cost_patch.assert_called_once_with(
        version_patches.beam_cost_addresses,
        game_patches.game_specific, dol_file
    )
    mock_apply_starting_visor_patch.assert_called_once_with(
        version_patches.starting_beam_visor,
        {"foo": "bar"}, dol_file,
    )


def test_get_dol_path():
    assert dol_patcher._get_dol_path(Path("foo")) == Path("foo", "sys", "main.dol")




from typing import Iterator

from randovania.game_description.assignment import PickupAssignment
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import Node, PickupNode


def filter_pickup_nodes(nodes: Iterator[Node]) -> Iterator[PickupNode]:
    for node in nodes:
        if isinstance(node, PickupNode):
            yield node


def filter_unassigned_pickup_nodes(nodes: Iterator[Node],
                                   pickup_assignment: PickupAssignment,
                                   ) -> Iterator[PickupNode]:
    for node in filter_pickup_nodes(nodes):
        if node.pickup_index not in pickup_assignment:
            yield node


class UnableToGenerate(RuntimeError):
    pass


def should_have_hint(item_category: ItemCategory) -> bool:
    return item_category.is_major_category



from typing import List, NamedTuple, Dict, TypeVar

from randovania.game_description.resources.damage_resource_info import DamageResourceInfo
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo


class MissingResource(ValueError):
    pass


T = TypeVar("T")


def find_resource_info_with_id(info_list: List[T], index: int, resource_type: ResourceType) -> T:
    for info in info_list:
        if info.index == index:
            return info
    indices = [info.index for info in info_list]
    raise MissingResource(f"{resource_type} Resource with index {index} not found in {indices}")


def find_resource_info_with_long_name(info_list: List[T], long_name: str) -> T:
    for info in info_list:
        if info.long_name == long_name:
            return info
    raise MissingResource(f"Resource with long_name '{long_name}' not found in {len(info_list)} resources")


class ResourceDatabase(NamedTuple):
    item: List[ItemResourceInfo]
    event: List[SimpleResourceInfo]
    trick: List[TrickResourceInfo]
    damage: List[DamageResourceInfo]
    version: List[SimpleResourceInfo]
    misc: List[SimpleResourceInfo]
    requirement_template: Dict[str, "Requirement"]
    energy_tank_item_index: int
    item_percentage_index: int
    multiworld_magic_item_index: int

    def get_by_type(self, resource_type: ResourceType) -> List[ResourceInfo]:
        if resource_type == ResourceType.ITEM:
            return self.item
        elif resource_type == ResourceType.EVENT:
            return self.event
        elif resource_type == ResourceType.TRICK:
            return self.trick
        elif resource_type == ResourceType.DAMAGE:
            return self.damage
        elif resource_type == ResourceType.VERSION:
            return self.version
        elif resource_type == ResourceType.MISC:
            return self.misc
        else:
            raise ValueError(
                "Invalid resource_type: {}".format(resource_type))

    def get_by_type_and_index(self, resource_type: ResourceType,
                              index: int) -> ResourceInfo:
        return find_resource_info_with_id(self.get_by_type(resource_type), index, resource_type)

    def get_item(self, index: int) -> ItemResourceInfo:
        return self.get_by_type_and_index(ResourceType.ITEM, index)

    @property
    def item_percentage(self) -> ItemResourceInfo:
        return self.get_by_type_and_index(ResourceType.ITEM, self.item_percentage_index)

    @property
    def energy_tank(self) -> ItemResourceInfo:
        return self.get_by_type_and_index(ResourceType.ITEM, self.energy_tank_item_index)

    @property
    def multiworld_magic_item(self) -> ItemResourceInfo:
        return self.get_item(self.multiworld_magic_item_index)



import copy
import dataclasses
from enum import Enum
from typing import Dict, Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder, BitPackEnum
from randovania.game_description.resources.translator_gate import TranslatorGate


class LayoutTranslatorRequirement(BitPackEnum, Enum):
    VIOLET = "violet"
    AMBER = "amber"
    EMERALD = "emerald"
    COBALT = "cobalt"
    RANDOM = "random"
    REMOVED = "removed"
    RANDOM_WITH_REMOVED = "random-removed"

    @property
    def item_index(self) -> int:
        if self in (LayoutTranslatorRequirement.RANDOM, LayoutTranslatorRequirement.RANDOM_WITH_REMOVED):
            raise ValueError("The random Requirement shouldn't be used for item_index")
        return ITEM_INDICES[self]

    @property
    def long_name(self) -> str:
        return LONG_NAMES[self]


ITEM_INDICES = {
    LayoutTranslatorRequirement.VIOLET: 97,
    LayoutTranslatorRequirement.AMBER: 98,
    LayoutTranslatorRequirement.EMERALD: 99,
    LayoutTranslatorRequirement.COBALT: 100,
    LayoutTranslatorRequirement.REMOVED: 9,
}

LONG_NAMES = {
    LayoutTranslatorRequirement.VIOLET: "Violet Translator",
    LayoutTranslatorRequirement.AMBER: "Amber Translator",
    LayoutTranslatorRequirement.EMERALD: "Emerald Translator",
    LayoutTranslatorRequirement.COBALT: "Cobalt Translator",
    LayoutTranslatorRequirement.RANDOM: "Random",
    LayoutTranslatorRequirement.REMOVED: "Unlocked",
    LayoutTranslatorRequirement.RANDOM_WITH_REMOVED: "Random with Unlocked",
}


@dataclasses.dataclass(frozen=True)
class TranslatorConfiguration(BitPackValue):
    translator_requirement: Dict[TranslatorGate, LayoutTranslatorRequirement]
    fixed_gfmc_compound: bool = True
    fixed_torvus_temple: bool = True
    fixed_great_temple: bool = True

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        from randovania.layout import configuration_factory
        templates = [
            configuration_factory.get_vanilla_actual_translator_configurations(),
            configuration_factory.get_vanilla_colors_translator_configurations(),
            self.with_full_random().translator_requirement,
            self.translator_requirement,
        ]
        yield from bitpacking.pack_array_element(self.translator_requirement, templates)
        if templates.index(self.translator_requirement) == 3:
            for translator in self.translator_requirement.values():
                yield from translator.bit_pack_encode({})

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "TranslatorConfiguration":
        from randovania.layout import configuration_factory
        templates = [
            configuration_factory.get_vanilla_actual_translator_configurations(),
            configuration_factory.get_vanilla_colors_translator_configurations(),
            cls.default().with_full_random().translator_requirement,
            None,
        ]
        translator_requirement = decoder.decode_element(templates)
        if translator_requirement is None:
            translator_requirement = {}
            for gate in templates[0].keys():
                translator_requirement[gate] = LayoutTranslatorRequirement.bit_pack_unpack(decoder, {})

        return cls(
            translator_requirement,
        )

    @property
    def as_json(self) -> dict:
        return {
            "translator_requirement": {
                str(key.index): item.value
                for key, item in self.translator_requirement.items()
            },
        }

    @classmethod
    def from_json(cls, value: dict) -> "TranslatorConfiguration":
        default = cls.default()

        params = copy.copy(value)

        translator_requirement = copy.copy(default.translator_requirement)
        for key, item in params.pop("translator_requirement").items():
            translator_requirement[TranslatorGate(int(key))] = LayoutTranslatorRequirement(item)

        return cls(translator_requirement, **params)

    @classmethod
    def default(cls) -> "TranslatorConfiguration":
        from randovania.layout import configuration_factory
        return cls(configuration_factory.get_vanilla_actual_translator_configurations())

    def with_vanilla_actual(self):
        from randovania.layout import configuration_factory

        result = copy.copy(self)
        return dataclasses.replace(
            result,
            translator_requirement=configuration_factory.get_vanilla_actual_translator_configurations())

    def with_vanilla_colors(self) -> "TranslatorConfiguration":
        from randovania.layout import configuration_factory

        result = copy.copy(self)
        return dataclasses.replace(
            result,
            translator_requirement=configuration_factory.get_vanilla_colors_translator_configurations())

    def with_full_random(self) -> "TranslatorConfiguration":
        result = copy.copy(self)
        return dataclasses.replace(result,
                                   translator_requirement={
                                       key: LayoutTranslatorRequirement.RANDOM
                                       for key in self.translator_requirement.keys()
                                   })

    def replace_requirement_for_gate(self, gate: TranslatorGate,
                                     requirement: LayoutTranslatorRequirement,
                                     ) -> "TranslatorConfiguration":
        """
        Replaces the requirement for the given gate. The gate must already have a requirement.
        :param gate:
        :param requirement:
        :return:
        """
        assert gate in self.translator_requirement
        result = copy.copy(self)

        new_requirement = copy.copy(self.translator_requirement)
        new_requirement[gate] = requirement

        return dataclasses.replace(result, translator_requirement=new_requirement)



import asyncio

from PySide2.QtWidgets import QDialog

from randovania.gui.generated.background_process_dialog_ui import Ui_BackgroundProcessDialog
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin


class BackgroundProcessDialog(QDialog, BackgroundTaskMixin, Ui_BackgroundProcessDialog):
    def __init__(self):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self.background_tasks_button_lock_signal.connect(self.update_button_label)
        self.progress_update_signal.connect(self.update_progress)
        self.update_button.clicked.connect(self.on_button_press)

    def on_button_press(self):
        if self.has_background_process:
            self.stop_background_process()
        else:
            self.accept()

    def update_button_label(self, value: bool):
        self.update_button.setText("Close" if value else "Cancel")

    def update_progress(self, message: str, percentage: int):
        self.progress_label.setText(message)
        if "Aborted" in message:
            percentage = 0
        if percentage >= 0:
            self.progress_bar.setRange(0, 100)
            self.progress_bar.setValue(percentage)
        else:
            self.progress_bar.setRange(0, 0)

    async def run_in_background_async_then_close(self, *args, **kwargs):
        result = await self.run_in_background_async(*args, **kwargs)
        self.accept()
        return result

    @classmethod
    async def open_for_background_task(cls, target, starting_message: str):
        dialog = cls()
        result = await asyncio.gather(
            dialog.run_in_background_async_then_close(target, starting_message),
            async_dialog.execute_dialog(dialog)
        )
        return result[0]



import copy
from typing import Optional, Tuple, Iterator

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import ResourceNode, Node
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo, CurrentResources, \
    add_resource_gain_to_current_resources, add_resources_into_another, convert_resource_gain_to_current_resources
from randovania.game_description.world_list import WorldList


def _energy_tank_difference(new_resources: CurrentResources,
                            old_resources: CurrentResources,
                            database: ResourceDatabase,
                            ) -> int:
    return new_resources.get(database.energy_tank, 0) - old_resources.get(database.energy_tank, 0)


class State:
    resources: CurrentResources
    collected_resource_nodes: Tuple[ResourceNode, ...]
    energy: int
    node: Node
    patches: GamePatches
    previous_state: Optional["State"]
    path_from_previous_state: Tuple[Node, ...]
    resource_database: ResourceDatabase
    world_list: WorldList

    def __init__(self,
                 resources: CurrentResources,
                 collected_resource_nodes: Tuple[ResourceNode, ...],
                 energy: int,
                 node: Node,
                 patches: GamePatches,
                 previous: Optional["State"],
                 resource_database: ResourceDatabase,
                 world_list: WorldList):

        self.resources = resources
        self.collected_resource_nodes = collected_resource_nodes
        self.node = node
        self.patches = patches
        self.path_from_previous_state = ()
        self.previous_state = previous
        self.resource_database = resource_database
        self.world_list = world_list

        # We place this last because we need resource_database set
        self.energy = min(energy, self.maximum_energy)

    def has_resource(self, resource: ResourceInfo) -> bool:
        return self.resources.get(resource, 0) > 0

    def copy(self) -> "State":
        return State(copy.copy(self.resources),
                     self.collected_resource_nodes,
                     self.energy,
                     self.node,
                     self.patches,
                     self.previous_state,
                     self.resource_database,
                     self.world_list)

    @property
    def collected_pickup_indices(self) -> Iterator[PickupIndex]:
        for resource, count in self.resources.items():
            if isinstance(resource, PickupIndex) and count > 0:
                yield resource

    @property
    def collected_scan_assets(self) -> Iterator[LogbookAsset]:
        for resource, count in self.resources.items():
            if isinstance(resource, LogbookAsset) and count > 0:
                yield resource

    def take_damage(self, damage: int) -> "State":
        return State(self.resources, self.collected_resource_nodes, self.energy - damage, self.node, self.patches, self,
                     self.resource_database, self.world_list)

    def heal(self) -> "State":
        return State(self.resources, self.collected_resource_nodes, self.maximum_energy, self.node, self.patches, self,
                     self.resource_database, self.world_list)

    def _energy_for(self, resources: CurrentResources) -> int:
        energy_per_tank = self.patches.game_specific.energy_per_tank if self.patches.game_specific is not None else 100
        return (energy_per_tank - 1) + (energy_per_tank * resources.get(self.resource_database.energy_tank, 0))

    @property
    def maximum_energy(self) -> int:
        return self._energy_for(self.resources)

    def collect_resource_node(self, node: ResourceNode, new_energy: int) -> "State":
        """
        Creates a new State that has the given ResourceNode collected.
        :param node:
        :param new_energy: How much energy you should have when collecting this resource
        :return:
        """

        if not node.can_collect(self.patches, self.resources, self.world_list.all_nodes):
            raise ValueError(
                "Trying to collect an uncollectable node'{}'".format(node))

        new_resources = copy.copy(self.resources)
        add_resource_gain_to_current_resources(node.resource_gain_on_collect(self.patches, self.resources,
                                                                             self.world_list.all_nodes),
                                               new_resources)

        energy = new_energy
        if _energy_tank_difference(new_resources, self.resources, self.resource_database) > 0:
            energy = self._energy_for(new_resources)

        return State(new_resources, self.collected_resource_nodes + (node,), energy, self.node, self.patches, self,
                     self.resource_database, self.world_list)

    def act_on_node(self, node: ResourceNode, path: Tuple[Node, ...] = (), new_energy: Optional[int] = None) -> "State":
        if new_energy is None:
            new_energy = self.energy
        new_state = self.collect_resource_node(node, new_energy)
        new_state.node = node
        new_state.path_from_previous_state = path
        return new_state

    def assign_pickup_resources(self, pickup: PickupEntry) -> "State":
        new_resources = copy.copy(self.resources)
        add_resource_gain_to_current_resources(pickup.resource_gain(self.resources), new_resources)

        energy = self.energy
        if _energy_tank_difference(new_resources, self.resources, self.resource_database) > 0:
            energy = self._energy_for(new_resources)

        return State(
            new_resources,
            self.collected_resource_nodes,
            energy,
            self.node,
            self.patches,
            self,
            self.resource_database,
            self.world_list,
        )

    def assign_pickup_to_starting_items(self, pickup: PickupEntry) -> "State":
        pickup_resources = convert_resource_gain_to_current_resources(pickup.resource_gain(self.resources))

        # Make sure there's no item percentage on starting items
        pickup_resources.pop(self.resource_database.item_percentage, None)

        new_resources = copy.copy(self.resources)
        add_resources_into_another(new_resources, pickup_resources)
        new_patches = self.patches.assign_extra_initial_items(pickup_resources)

        if self.patches.game_specific is not None:
            energy_per_tank = self.patches.game_specific.energy_per_tank
        else:
            energy_per_tank = 100

        return State(
            new_resources,
            self.collected_resource_nodes,
            self.energy + _energy_tank_difference(new_resources, self.resources,
                                                  self.resource_database) * energy_per_tank,
            self.node,
            new_patches,
            self,
            self.resource_database,
            self.world_list,
        )


def add_pickup_to_state(state: State, pickup: PickupEntry):
    """
    Modifies inplace the given state, adding the resources of the given pickup
    :param state:
    :param pickup:
    :return:
    """
    add_resource_gain_to_current_resources(pickup.resource_gain(state.resources), state.resources)


def state_with_pickup(state: State,
                      pickup: PickupEntry,
                      ) -> State:
    """
    Returns a new State that follows the given State and also has the resource gain of the given pickup
    :param state:
    :param pickup:
    :return:
    """
    new_state = state.copy()
    new_state.previous_state = state
    add_pickup_to_state(new_state, pickup)
    if new_state.maximum_energy > state.maximum_energy:
        new_state.energy = new_state.maximum_energy
    return new_state



from dataclasses import dataclass
from typing import Optional, Tuple, Iterator

from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_info import ResourceGainTuple, ResourceGain, ResourceQuantity
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo


@dataclass(frozen=True)
class ConditionalResources:
    name: Optional[str]
    item: Optional[ItemResourceInfo]
    resources: ResourceGainTuple


@dataclass(frozen=True)
class ResourceConversion:
    source: ItemResourceInfo
    target: ItemResourceInfo
    clear_source: bool = True
    overwrite_target: bool = False


MAXIMUM_PICKUP_CONDITIONAL_RESOURCES = 3
MAXIMUM_PICKUP_RESOURCES = 8
MAXIMUM_PICKUP_CONVERSION = 2


@dataclass(frozen=True)
class PickupEntry:
    name: str
    model_index: int
    item_category: ItemCategory
    broad_category: ItemCategory
    resources: Tuple[ConditionalResources, ...]
    convert_resources: Tuple[ResourceConversion, ...] = tuple()
    probability_offset: float = 0
    probability_multiplier: float = 1

    def __post_init__(self):
        if not isinstance(self.resources, tuple):
            raise ValueError("resources should be a tuple, got {}".format(self.resources))

        if len(self.resources) < 1:
            raise ValueError("resources should have at least 1 value")

        if len(self.resources) > MAXIMUM_PICKUP_CONDITIONAL_RESOURCES:
            raise ValueError(f"resources should have at most {MAXIMUM_PICKUP_CONDITIONAL_RESOURCES} "
                             f"values, got {len(self.resources)}")

        for i, conditional in enumerate(self.resources):
            if not isinstance(conditional, ConditionalResources):
                raise ValueError(f"Conditional at {i} should be a ConditionalResources")

            if len(conditional.resources) > MAXIMUM_PICKUP_RESOURCES:
                raise ValueError(f"Conditional at {i} should have at most {MAXIMUM_PICKUP_RESOURCES} "
                                 f"resources, got {len(conditional.resources)}")

            if i == 0:
                if conditional.item is not None:
                    raise ValueError("Conditional at 0 should not have a condition")
            else:
                if conditional.item is None:
                    raise ValueError(f"Conditional at {i} should have a condition")

        if len(self.convert_resources) > MAXIMUM_PICKUP_CONVERSION:
            raise ValueError(f"convert_resources should have at most {MAXIMUM_PICKUP_CONVERSION} value")

        for i, conversion in enumerate(self.convert_resources):
            if not conversion.clear_source or conversion.overwrite_target:
                raise ValueError(f"clear_source and overwrite_target should be True and False, "
                                 f"got {conversion.clear_source} and {conversion.overwrite_target} for index {i}")

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        return self.name < other.name

    def conditional_for_resources(self, current_resources) -> ConditionalResources:
        last_conditional: Optional[ConditionalResources] = None

        for conditional in self.resources:
            if conditional.item is None or current_resources.get(conditional.item, 0) > 0:
                last_conditional = conditional
            else:
                break

        assert last_conditional is not None
        return last_conditional

    def conversion_resource_gain(self, current_resources):
        for conversion in self.convert_resources:
            quantity = current_resources.get(conversion.source, 0)
            yield conversion.source, -quantity
            yield conversion.target, quantity

    def resource_gain(self, current_resources) -> ResourceGain:
        yield from self.conditional_for_resources(current_resources).resources
        yield from self.conversion_resource_gain(current_resources)

    def __str__(self):
        return "Pickup {}".format(self.name)

    @property
    def all_resources(self) -> Iterator[ResourceQuantity]:
        for conditional in self.resources:
            yield from conditional.resources

    @property
    def is_expansion(self) -> bool:
        return self.item_category == ItemCategory.EXPANSION



import dataclasses
from unittest.mock import MagicMock

import pytest

from randovania.game_description.area import Area
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.hint import Hint, HintType, HintLocationPrecision, HintItemPrecision, PrecisionPair, \
    RelativeDataItem, RelativeDataArea, HintRelativeAreaName, HintDarkTemple
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import LogbookNode, PickupNode
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world import World
from randovania.game_description.world_list import WorldList
from randovania.games.prime import echoes_items
from randovania.games.prime.patcher_file_lib import item_hints
from randovania.games.prime.patcher_file_lib.hint_name_creator import LocationHintCreator
from randovania.games.prime.patcher_file_lib.item_hints import RelativeItemFormatter, RelativeAreaFormatter
from randovania.interface_common.players_configuration import PlayersConfiguration


@pytest.fixture(name="pickup")
def _pickup() -> PickupEntry:
    return PickupEntry(
        name="Pickup",
        model_index=0,
        item_category=ItemCategory.MOVEMENT,
        broad_category=ItemCategory.LIFE_SUPPORT,
        resources=(
            ConditionalResources(None, None, ()),
        ),
    )


@pytest.fixture(name="players_config")
def _players_configuration() -> PlayersConfiguration:
    return PlayersConfiguration(
        player_index=0,
        player_names={0: "Player 1"},
    )


def _create_world_list(asset_id: int, pickup_index: PickupIndex):
    logbook_node = LogbookNode("Logbook A", True, None, 0, asset_id, None, None, None, None)
    pickup_node = PickupNode("Pickup Node", True, None, 1, pickup_index, True)

    world_list = WorldList([
        World("World", "Dark World", 5000, [
            Area("Area", False, 10000, 0, True, [logbook_node, pickup_node], {}),
            Area("Other Area", False, 20000, 0, True, [PickupNode(f"Pickup {i}", True, None, 1, PickupIndex(i), True)
                                                       for i in range(pickup_index.index)], {}),
        ]),
    ])

    return logbook_node, pickup_node, world_list


def test_create_hints_nothing(empty_patches, players_config):
    # Setup
    asset_id = 1000
    pickup_index = PickupIndex(0)

    logbook_node, _, world_list = _create_world_list(asset_id, pickup_index)

    patches = dataclasses.replace(
        empty_patches,
        hints={
            logbook_node.resource(): Hint(HintType.LOCATION,
                                          PrecisionPair(HintLocationPrecision.DETAILED, HintItemPrecision.DETAILED,
                                                        include_owner=False),
                                          pickup_index)
        })
    rng = MagicMock()

    # Run
    result = item_hints.create_hints({0: patches}, players_config, world_list, rng)

    # Assert
    message = ("The &push;&main-color=#FF6705B3;Energy Transfer Module&pop; can be found in "
               "&push;&main-color=#FF3333;World - Area&pop;.")
    assert result == [
        {'asset_id': asset_id, 'strings': [message, '', message]}
    ]


def test_create_hints_item_joke(empty_patches, players_config, pickup):
    # Setup
    asset_id = 1000
    logbook_node, _, world_list = _create_world_list(asset_id, PickupIndex(50))

    patches = dataclasses.replace(
        empty_patches,
        hints={
            logbook_node.resource(): Hint(HintType.JOKE, None)
        })
    rng = MagicMock()

    # Run
    result = item_hints.create_hints({0: patches}, players_config, world_list, rng)

    # Assert
    joke = "While walking, holding L makes you move faster."
    message = f"&push;&main-color=#45F731;{joke}&pop;"
    assert result[0]['strings'][0] == message
    assert result == [{'asset_id': asset_id, 'strings': [message, '', message]}]


@pytest.mark.parametrize(["indices", "expected_message"], [
    ((0, 1, 2), "The keys to &push;&main-color=#FF6705B3;Dark Torvus Temple&pop; can "
                "all be found in &push;&main-color=#FF3333;Temple Grounds&pop;."),
    ((0, 1, 118), "The keys to &push;&main-color=#FF6705B3;Dark Torvus Temple&pop; can "
                  "be found in &push;&main-color=#FF3333;Sanctuary Fortress&pop; "
                  "and &push;&main-color=#FF3333;Temple Grounds&pop;."),
    ((0, 41, 118), "The keys to &push;&main-color=#FF6705B3;Dark Torvus Temple&pop; can "
                   "be found in &push;&main-color=#FF3333;Dark Agon Wastes&pop;, "
                   "&push;&main-color=#FF3333;Sanctuary Fortress&pop; and "
                   "&push;&main-color=#FF3333;Temple Grounds&pop;."),
])
def test_create_hints_item_dark_temple_keys(empty_patches, players_config, echoes_game_description, pickup,
                                            indices, expected_message):
    # Setup
    db = echoes_game_description.resource_database
    keys = [
        (
            PickupIndex(index),
            dataclasses.replace(pickup, resources=(
                ConditionalResources(None, None, ((db.get_item(item), 1),)),
            ))
        )
        for index, item in zip(indices, echoes_items.DARK_TEMPLE_KEY_ITEMS[1])
    ]

    patches = dataclasses.replace(
        empty_patches,
        pickup_assignment={
            pickup_index: PickupTarget(key, 0)
            for pickup_index, key in keys
        })

    hint = Hint(HintType.RED_TEMPLE_KEY_SET, None, dark_temple=HintDarkTemple.TORVUS_BOG)

    # Run
    result = item_hints.create_message_for_hint(hint, {0: patches}, players_config, None, {},
                                                echoes_game_description.world_list)

    # Assert
    assert result == expected_message


def test_create_message_for_hint_dark_temple_no_keys(empty_patches, players_config, echoes_game_description):
    # Setup
    hint = Hint(HintType.RED_TEMPLE_KEY_SET, None, dark_temple=HintDarkTemple.TORVUS_BOG)

    # Run
    result = item_hints.create_message_for_hint(hint, {0: empty_patches}, players_config, None, {},
                                                echoes_game_description.world_list)

    # Assert
    assert result == 'The keys to &push;&main-color=#FF6705B3;Dark Torvus Temple&pop; are nowhere to be found.'


@pytest.mark.parametrize("item", [
    (HintItemPrecision.DETAILED, "The", "&push;&main-color=#FF6705B3;Pickup&pop;"),
    (HintItemPrecision.PRECISE_CATEGORY, "A", "&push;&main-color=#FF6705B3;movement system&pop;"),
    (HintItemPrecision.GENERAL_CATEGORY, "A", "&push;&main-color=#FF6705B3;major upgrade&pop;"),
    (HintItemPrecision.BROAD_CATEGORY, "A", "&push;&main-color=#FF6705B3;life support system&pop;"),
])
@pytest.mark.parametrize("location", [
    (HintLocationPrecision.DETAILED, "&push;&main-color=#FF3333;World - Area&pop;"),
    (HintLocationPrecision.WORLD_ONLY, "&push;&main-color=#FF3333;World&pop;"),
])
@pytest.mark.parametrize("owner", [False, True])
@pytest.mark.parametrize("is_multiworld", [False, True])
def test_create_hints_item_location(empty_patches, pickup, item, location, owner, is_multiworld):
    # Setup
    asset_id = 1000
    pickup_index = PickupIndex(50)
    logbook_node, _, world_list = _create_world_list(asset_id, pickup_index)
    players_config = PlayersConfiguration(
        player_index=0,
        player_names={i: f"Player {i +  1}"
                      for i in range(int(is_multiworld) + 1)},
    )
    location_precision, determiner, item_name = item
    if owner and is_multiworld:
        determiner = "Player 1's"

    patches = dataclasses.replace(
        empty_patches,
        pickup_assignment={
            pickup_index: PickupTarget(pickup, 0),
        },
        hints={
            logbook_node.resource(): Hint(HintType.LOCATION,
                                          PrecisionPair(location[0], location_precision, include_owner=owner),
                                          pickup_index)
        })
    rng = MagicMock()

    # Run
    result = item_hints.create_hints({0: patches}, players_config, world_list, rng)

    # Assert
    message = "{} {} can be found in {}.".format(determiner, item_name, location[1])
    # message = "The Flying Ing Cache in {} contains {}.".format(location[1], item[1])
    assert result[0]['strings'][0] == message
    assert result == [{'asset_id': asset_id, 'strings': [message, '', message]}]


@pytest.mark.parametrize("pickup_index_and_guardian", [
    (PickupIndex(43), "&push;&main-color=#FF3333;Amorbis&pop;"),
    (PickupIndex(79), "&push;&main-color=#FF3333;Chykka&pop;"),
    (PickupIndex(115), "&push;&main-color=#FF3333;Quadraxis&pop;"),
])
@pytest.mark.parametrize("item", [
    (HintItemPrecision.DETAILED, "the &push;&main-color=#FF6705B3;Pickup&pop;"),
    (HintItemPrecision.PRECISE_CATEGORY, "a &push;&main-color=#FF6705B3;movement system&pop;"),
    (HintItemPrecision.GENERAL_CATEGORY, "a &push;&main-color=#FF6705B3;major upgrade&pop;"),
    (HintItemPrecision.BROAD_CATEGORY, "a &push;&main-color=#FF6705B3;life support system&pop;"),
])
def test_create_hints_guardians(empty_patches, pickup_index_and_guardian, pickup, item, players_config):
    # Setup
    asset_id = 1000
    pickup_index, guardian = pickup_index_and_guardian

    logbook_node, _, world_list = _create_world_list(asset_id, pickup_index)

    patches = dataclasses.replace(
        empty_patches,
        pickup_assignment={
            pickup_index: PickupTarget(pickup, 0),
        },
        hints={
            logbook_node.resource(): Hint(HintType.LOCATION,
                                          PrecisionPair(HintLocationPrecision.GUARDIAN, item[0],
                                                        include_owner=False),
                                          pickup_index)
        })
    rng = MagicMock()

    # Run
    result = item_hints.create_hints({0: patches}, players_config, world_list, rng)

    # Assert
    message = f"{guardian} is guarding {item[1]}."
    assert result == [{'asset_id': asset_id, 'strings': [message, '', message]}]


@pytest.mark.parametrize("item", [
    (HintItemPrecision.DETAILED, "the &push;&main-color=#FF6705B3;Pickup&pop;"),
    (HintItemPrecision.PRECISE_CATEGORY, "a &push;&main-color=#FF6705B3;movement system&pop;"),
    (HintItemPrecision.GENERAL_CATEGORY, "a &push;&main-color=#FF6705B3;major upgrade&pop;"),
    (HintItemPrecision.BROAD_CATEGORY, "a &push;&main-color=#FF6705B3;life support system&pop;"),
])
def test_create_hints_light_suit_location(empty_patches, players_config, pickup, item):
    # Setup
    asset_id = 1000
    pickup_index = PickupIndex(50)

    logbook_node, _, world_list = _create_world_list(asset_id, pickup_index)

    patches = dataclasses.replace(
        empty_patches,
        pickup_assignment={
            pickup_index: PickupTarget(pickup, 0),
        },
        hints={
            logbook_node.resource(): Hint(HintType.LOCATION,
                                          PrecisionPair(HintLocationPrecision.LIGHT_SUIT_LOCATION, item[0],
                                                        include_owner=False),
                                          pickup_index)
        })
    rng = MagicMock()

    # Run
    result = item_hints.create_hints({0: patches}, players_config, world_list, rng)

    # Assert
    message = f"U-Mos's reward for returning the Sanctuary energy is {item[1]}."
    assert result == [{'asset_id': asset_id, 'strings': [message, '', message]}]


@pytest.mark.parametrize(["reference_precision", "reference_name"], [
    (HintItemPrecision.DETAILED, "the Reference Pickup"),
    (HintItemPrecision.PRECISE_CATEGORY, "a movement system"),
    (HintItemPrecision.BROAD_CATEGORY, "a life support system"),
])
@pytest.mark.parametrize(["distance_precise", "distance_text"], [
    (False, "up to"),
    (True, "exactly"),
])
def test_create_message_for_hint_relative_item(echoes_game_description, pickup, players_config,
                                               distance_precise, distance_text,
                                               reference_precision, reference_name):
    world_list = echoes_game_description.world_list
    patches = echoes_game_description.create_game_patches().assign_pickup_assignment({
        PickupIndex(5): PickupTarget(pickup, 0),
        PickupIndex(15): PickupTarget(dataclasses.replace(pickup, name="Reference Pickup"), 0),
    })

    hint_name_creator = LocationHintCreator(world_list, None, None)
    location_formatters = {
        HintLocationPrecision.RELATIVE_TO_INDEX: RelativeItemFormatter(world_list, patches, players_config),
    }
    hint = Hint(
        HintType.LOCATION,
        PrecisionPair(HintLocationPrecision.RELATIVE_TO_INDEX, HintItemPrecision.DETAILED, include_owner=False,
                      relative=RelativeDataItem(distance_precise, PickupIndex(15), reference_precision)),
        PickupIndex(5)
    )

    # Run
    result = item_hints.create_message_for_hint(hint, {0: patches}, players_config, hint_name_creator,
                                                location_formatters,
                                                world_list)

    # Assert
    assert result == (f'The &push;&main-color=#FF6705B3;Pickup&pop; can be found '
                      f'&push;&main-color=#FF3333;{distance_text} 7 rooms&pop; away from {reference_name}.')


@pytest.mark.parametrize(["distance_precise", "distance_text"], [
    (False, "up to"),
    (True, "exactly"),
])
def test_create_message_for_hint_relative_area(echoes_game_description, pickup, players_config,
                                               distance_precise, distance_text):
    world_list = echoes_game_description.world_list
    patches = echoes_game_description.create_game_patches().assign_pickup_assignment({
        PickupIndex(5): PickupTarget(pickup, 0),
    })

    hint_name_creator = LocationHintCreator(world_list, None, None)
    location_formatters = {HintLocationPrecision.RELATIVE_TO_AREA: RelativeAreaFormatter(world_list, patches)}
    hint = Hint(
        HintType.LOCATION,
        PrecisionPair(HintLocationPrecision.RELATIVE_TO_AREA, HintItemPrecision.DETAILED, include_owner=False,
                      relative=RelativeDataArea(distance_precise,
                                                AreaLocation(1039999561, 3822429534),
                                                HintRelativeAreaName.NAME)),
        PickupIndex(5)
    )

    # Run
    result = item_hints.create_message_for_hint(hint, {0: patches}, players_config,
                                                hint_name_creator, location_formatters,
                                                world_list)

    # Assert
    assert result == (f'The &push;&main-color=#FF6705B3;Pickup&pop; can be found '
                      f'&push;&main-color=#FF3333;{distance_text} 10 rooms&pop; away from Torvus Bog - Great Bridge.')



import copy
import dataclasses
import json
from pathlib import Path

import pytest

from randovania.gui import tracker_window
from randovania.layout.elevators import LayoutElevators
from randovania.layout.translator_configuration import LayoutTranslatorRequirement


@pytest.fixture(params=[{},
                        {"elevators": LayoutElevators.ONE_WAY_ANYTHING,
                         "translator_configuration": True}],
                name="layout_config")
def _layout_config(request, default_layout_configuration):
    if "translator_configuration" in request.param:
        translator_requirement = copy.copy(default_layout_configuration.translator_configuration.translator_requirement)
        for gate in translator_requirement.keys():
            translator_requirement[gate] = LayoutTranslatorRequirement.RANDOM
            break

        new_gate = dataclasses.replace(default_layout_configuration.translator_configuration,
                                       translator_requirement=translator_requirement)
        request.param["translator_configuration"] = new_gate
    return dataclasses.replace(default_layout_configuration, **request.param)


def test_load_previous_state_no_previous_layout(tmp_path: Path, default_layout_configuration):
    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_previous_layout_not_json(tmp_path: Path, default_layout_configuration):
    # Setup
    tmp_path.joinpath("layout_configuration.json").write_text("this is not a json")

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_previous_layout_not_layout(tmp_path: Path, default_layout_configuration):
    # Setup
    tmp_path.joinpath("layout_configuration.json").write_text(json.dumps({"trick_level": "foo"}))
    tmp_path.joinpath("state.json").write_text("[]")

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_missing_state(tmp_path: Path, default_layout_configuration):
    # Setup
    tmp_path.joinpath("layout_configuration.json").write_text(json.dumps(default_layout_configuration.as_json))

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_invalid_state(tmp_path: Path, default_layout_configuration):
    # Setup
    tmp_path.joinpath("layout_configuration.json").write_text(json.dumps(default_layout_configuration.as_json))
    tmp_path.joinpath("state.json").write_text("")

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_success(tmp_path: Path, default_layout_configuration):
    # Setup
    data = {"asdf": 5, "zxcv": 123}
    tmp_path.joinpath("layout_configuration.json").write_text(json.dumps(default_layout_configuration.as_json))
    tmp_path.joinpath("state.json").write_text(json.dumps(data))

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result == data


@pytest.mark.parametrize("shuffle_advanced", [False, True])
def test_apply_previous_state(skip_qtbot, tmp_path: Path, default_layout_configuration, shuffle_advanced):
    if shuffle_advanced:
        translator_requirement = copy.copy(
            default_layout_configuration.translator_configuration.translator_requirement)
        for gate in translator_requirement.keys():
            translator_requirement[gate] = LayoutTranslatorRequirement.RANDOM
            break

        new_gate = dataclasses.replace(default_layout_configuration.translator_configuration,
                                       translator_requirement=translator_requirement)
        layout_config = dataclasses.replace(
            default_layout_configuration,
            elevators=LayoutElevators.ONE_WAY_ANYTHING,
            translator_configuration=new_gate)
    else:
        layout_config = default_layout_configuration

    state = {
        "actions": [
            0,
            4
        ],
        "collected_pickups": {
            'Amber Translator': 0,
            'Annihilator Beam': 0,
            'Boost Ball': 0,
            'Cobalt Translator': 0,
            'Dark Agon Key 1': 0,
            'Dark Agon Key 2': 0,
            'Dark Agon Key 3': 0,
            'Dark Ammo Expansion': 0,
            'Dark Beam': 0,
            'Dark Torvus Key 1': 0,
            'Dark Torvus Key 2': 0,
            'Dark Torvus Key 3': 0,
            'Dark Visor': 0,
            'Darkburst': 0,
            'Echo Visor': 0,
            'Emerald Translator': 0,
            'Energy Tank': 0,
            'Grapple Beam': 0,
            'Gravity Boost': 0,
            'Ing Hive Key 1': 0,
            'Ing Hive Key 2': 0,
            'Ing Hive Key 3': 0,
            'Light Ammo Expansion': 0,
            'Light Beam': 0,
            'Missile Expansion': 0,
            'Missile Launcher': 0,
            'Morph Ball Bomb': 0,
            'Power Bomb': 0,
            'Power Bomb Expansion': 0,
            'Progressive Suit': 0,
            'Screw Attack': 0,
            'Seeker Launcher': 0,
            'Sky Temple Key 1': 0,
            'Sky Temple Key 2': 0,
            'Sky Temple Key 3': 0,
            'Sky Temple Key 4': 0,
            'Sky Temple Key 5': 0,
            'Sky Temple Key 6': 0,
            'Sky Temple Key 7': 0,
            'Sky Temple Key 8': 0,
            'Sky Temple Key 9': 0,
            'Sonic Boom': 0,
            'Space Jump Boots': 1,
            'Spider Ball': 0,
            'Sunburst': 0,
            'Super Missile': 0,
            'Violet Translator': 0,
        },
        "elevators": {
            '122': None,
            '1245307': None,
            '1245332': None,
            '129': None,
            '152': None,
            '1572998': None,
            '1638535': None,
            '1966093': None,
            '2097251': None,
            '2162826': None,
            '2949235': None,
            '3342446': None,
            '3538975': None,
            '38': None,
            '393260': None,
            '4522032': None,
            '524321': None,
            '589851': None,
        },
        "translator_gates": {
            "0": None,
            '1': None,
            '10': None,
            '11': None,
            '12': None,
            '13': None,
            '14': None,
            '15': None,
            '16': None,
            '2': None,
            '3': None,
            '4': None,
            '5': None,
            '6': None,
            '7': None,
            '8': None,
            '9': None,
        },
        "starting_location": {
            "world_asset_id": 1006255871,
            "area_asset_id": 1655756413
        }
    }
    if shuffle_advanced:
        state["elevators"]['2949235'] = {'area_asset_id': 50083607, 'world_asset_id': 1119434212}
        state["translator_gates"]['0'] = 97
    tmp_path.joinpath("layout_configuration.json").write_text(json.dumps(layout_config.as_json), "utf-8")
    tmp_path.joinpath("state.json").write_text(json.dumps(state), "utf-8")

    # Run
    window = tracker_window.TrackerWindow(tmp_path, layout_config)
    skip_qtbot.add_widget(window)

    # Assert
    assert window.state_for_current_configuration() is not None
    persisted_data = json.loads(tmp_path.joinpath("state.json").read_text("utf-8"))
    assert persisted_data == state

    window.reset()
    window.persist_current_state()

    persisted_data = json.loads(tmp_path.joinpath("state.json").read_text("utf-8"))
    assert persisted_data != state
